import re
from tkinter import Tk

from subsearch.globals.constants import APP_PATHS


def get_sprites() -> dict:
    pattern = r"\b(\w+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\b"
    data = {}
    file_path = APP_PATHS.gui_styles / "sprites.tcl"
    with file_path.open() as file:
        content = file.read()
        matches = re.findall(pattern, content)

        for match in matches:
            key = match[0]
            values = tuple(map(int, match[1:]))
            data[key] = values

    return data


root = Tk(className="Subsearch")
sprites = get_sprites()



import tkinter as tk
from tkinter import ttk

from subsearch.globals import log
from subsearch.globals.constants import FILE_PATHS
from subsearch.gui import common_utils
from subsearch.gui.resources import config as cfg
from subsearch.utils import io_toml, string_parser


class Providers(ttk.Labelframe):
    def __init__(self, parent) -> None:
        ttk.Labelframe.__init__(self, parent)
        self.configure(text="Subtitle providers", padding=10)
        self.data = io_toml.load_toml_data(FILE_PATHS.config)
        self.chekboxes = {}
        self.last_key = ""

        self.provider_options: dict = {
            "opensubtitles_site": "Opensubtitles",
            "opensubtitles_hash": "Opensubtitles with hash",
            "yifysubtitles_site": "Yifysubtitles",
            "subsource_site": "Subsource",
        }
        for name, description in self.provider_options.items():
            self.provider_options[name] = [
                io_toml.load_toml_value(FILE_PATHS.config, "providers")[name],
                description,
            ]

        frame = None
        for enum, (key, value) in enumerate(self.provider_options.items()):
            if enum % 4 == 0:
                frame = tk.Frame(self)
                frame.pack(side=tk.LEFT, anchor="nw", expand=True)
            boolean = tk.BooleanVar()
            boolean.set(value[0])
            btn = ttk.Checkbutton(frame, text=value[1], onvalue=True, offvalue=False, variable=boolean)
            btn.pack(padx=4, pady=4, ipadx=20)
            self.chekboxes[btn] = key, boolean
            if self.data["providers"][key] is True:
                btn.configure(state="normal")

            btn.bind("<Enter>", self.enter_button)
            btn.bind("<Leave>", self.leave_button)

    def enter_button(self, event) -> None:
        btn = event.widget
        btn.bind("<ButtonPress-1>", self.press_button)

    def leave_button(self, event) -> None:
        btn = event.widget
        self.chekboxes[btn][0]

    def press_button(self, event) -> None:
        btn = event.widget
        btn.bind("<ButtonRelease-1>", self.toggle_types)

    def toggle_types(self, event) -> None:
        btn = event.widget
        key = self.chekboxes[btn][0]
        value = self.chekboxes[btn][1]
        if value.get() is True:
            self.data["providers"][key] = False
        elif value.get() is False:
            self.data["providers"][key] = True
        io_toml.dump_toml_data(FILE_PATHS.config, self.data)

    def toggle_providers(self, event) -> None:
        btn = event.widget
        key = btn["text"].replace(" ", "_").lower()
        if self.data["providers"][key] is True:
            self.data["providers"][key] = False
            btn.configure(fg=cfg.color.red)
        elif self.data["providers"][key] is False:
            self.data["providers"][key] = True
            btn.configure(fg=cfg.color.green)
        io_toml.dump_toml_data(FILE_PATHS.config, self.data)


class SubtitleFilters(ttk.Labelframe):
    def __init__(self, parent) -> None:
        ttk.Labelframe.__init__(self, parent)
        self.configure(text="Subtitle filters", padding=10)
        self.data = io_toml.load_toml_data(FILE_PATHS.config)
        self.subtitle_options: dict = {
            "subtitle_filters.hearing_impaired": "Hearing impaird",
            "subtitle_filters.non_hearing_impaired": "non-Hearing impaired",
            "subtitle_filters.only_foreign_parts": "Foreign parts only",
        }

        for name, description in self.subtitle_options.items():
            self.subtitle_options[name] = [io_toml.load_toml_value(FILE_PATHS.config, name), description]
        frame = None
        self.checkbuttons: dict[ttk.Checkbutton, tuple[str, tk.BooleanVar]] = {}
        for enum, (key, value) in enumerate(self.subtitle_options.items()):
            bool_value = value[0]
            description = value[1]
            if enum % 5 == 0:
                frame = ttk.Frame(self)
                frame.pack(side=tk.LEFT, anchor="nw", expand=True)

            boolean = tk.BooleanVar()
            boolean.set(bool_value)
            check_btn = ttk.Checkbutton(frame, text=description, onvalue=True, offvalue=False, variable=boolean)
            check_btn.pack(anchor="nw", padx=4, pady=4, ipadx=20)
            self.checkbuttons[check_btn] = key, boolean
            check_btn.bind("<Enter>", self.enter_button)

    def enter_button(self, event) -> None:
        btn = event.widget
        btn.bind("<ButtonPress-1>", self.press_button)

    def press_button(self, event) -> None:
        btn = event.widget
        btn.bind("<ButtonRelease-1>", self.toggle_types)

    def toggle_types(self, event) -> None:
        btn = event.widget
        key = self.checkbuttons[btn][0]
        value = self.checkbuttons[btn][1]
        if value.get() is True:
            io_toml.update_toml_key(FILE_PATHS.config, key, False)
        elif value.get() is False:
            io_toml.update_toml_key(FILE_PATHS.config, key, True)


class SubtitlePostProcessing(ttk.Labelframe):
    def __init__(self, parent) -> None:
        ttk.Labelframe.__init__(self, parent)
        self.configure(text="Subtitle post processing", padding=10)
        self.data = io_toml.load_toml_data(FILE_PATHS.config)

        self.subtitle_options: dict = {
            "subtitle_post_processing.rename": "Rename best subtitle",
            "subtitle_post_processing.move_best": "Move best subtitle",
            "subtitle_post_processing.move_all": "Move all subtitles",
        }
        for name, description in self.subtitle_options.items():
            self.subtitle_options[name] = [io_toml.load_toml_value(FILE_PATHS.config, name), description]
        frame = None
        self.checkbuttons: dict[ttk.Checkbutton, tuple[str, tk.BooleanVar]] = {}
        for enum, (key, value) in enumerate(self.subtitle_options.items()):
            bool_value = value[0]
            description = value[1]
            if enum % 1 == 0:
                frame = ttk.Frame(self)
                frame.pack(side=tk.LEFT, anchor="n")

            boolean = tk.BooleanVar()
            boolean.set(bool_value)
            check_btn = ttk.Checkbutton(frame, text=description, onvalue=True, offvalue=False, variable=boolean)
            check_btn.pack(padx=4, pady=4, ipadx=20)
            self.checkbuttons[check_btn] = key, boolean
            check_btn.bind("<Enter>", self.enter_button)

    def enter_button(self, event) -> None:
        btn = event.widget
        btn.bind("<ButtonPress-1>", self.press_button)

    def press_button(self, event) -> None:
        btn = event.widget
        btn.bind("<ButtonRelease-1>", self.toggle_buttons)

    def toggle_buttons(self, event) -> None:
        btn = event.widget
        key = self.checkbuttons[btn][0]
        value = self.checkbuttons[btn][1]
        if value.get() is True:
            io_toml.update_toml_key(FILE_PATHS.config, key, False)
        elif value.get() is False:
            io_toml.update_toml_key(FILE_PATHS.config, key, True)


class SubtitlePostProcessingDirectory(ttk.Labelframe):
    def __init__(self, parent) -> None:
        ttk.Labelframe.__init__(self, parent)
        self.configure(text="Subtitle move destination", padding=10)
        self.target_path = io_toml.load_toml_value(FILE_PATHS.config, "subtitle_post_processing.target_path")
        self.path_resolution = tk.StringVar()
        self._get_path_resolution()

        frame_l = tk.Frame(self)
        frame_c = tk.Frame(self)
        frame_r = tk.Frame(self)

        frame_l.pack(side=tk.LEFT, expand=False)
        frame_c.pack(side=tk.LEFT, expand=True, after=frame_l)
        frame_r.pack(side=tk.LEFT, expand=False, after=frame_c)

        self.btn_resolution = ttk.Button(frame_l, textvariable=self.path_resolution, width=10)
        self.entry_path = ttk.Entry(frame_c, width=50)
        apply_input = ttk.Button(frame_r, text="Apply", width=10)

        self.entry_path.insert(tk.END, f"{self.target_path}")

        self.btn_resolution.pack(anchor="center")
        self.entry_path.pack(anchor="center")
        apply_input.pack(anchor="center")

        self.btn_resolution.bind("<Enter>", self.enter_btn_resolution)
        self.btn_resolution.bind("<Leave>", self.leave_btn_resolution)
        apply_input.bind("<Enter>", self.enter_btn_apply_input)
        apply_input.bind("<Leave>", self.leave_btn_apply_input)
        self.entry_path.bind("<Enter>", self.enter_entry_path)

    def enter_entry_path(self, event) -> None:
        widget_ = event.widget
        path = widget_.get()
        old_path = path.replace("Invalid path: ", "")
        widget_.delete(0, tk.END)
        widget_.insert(tk.END, old_path)

    def enter_btn_apply_input(self, event) -> None:
        widget_ = event.widget
        path = self.verify_path()
        if self.entry_path.instate(["invalid"]):
            self.on_invalid_state(path)
            widget_.unbind("<ButtonPress-1>")
        elif not self.entry_path.instate(["invalid"]):
            widget_.bind("<ButtonPress-1>", self.update_config)

    def enter_btn_resolution(self, event) -> None:
        widget_ = event.widget
        widget_.bind("<ButtonPress-1>", self.toogle_path_resolution)

    def leave_btn_apply_input(self, event) -> None:
        widget_ = event.widget
        path = self.verify_path()
        if self.entry_path.instate(["invalid"]):
            self.on_invalid_state(path)
            widget_.unbind("<ButtonPress-1>")
        elif not self.entry_path.instate(["invalid"]):
            widget_.bind("<ButtonPress-1>", self.update_config)

    def leave_btn_resolution(self, event) -> None:
        widget_ = event.widget
        widget_.unbind("<ButtonPress-1>")

    def update_config(self, event) -> None:
        target_path = self.entry_path.get()
        path_resolution = self.path_resolution.get().lower()
        io_toml.update_toml_key(FILE_PATHS.config, "subtitle_post_processing.target_path", target_path)
        io_toml.update_toml_key(FILE_PATHS.config, "subtitle_post_processing.path_resolution", path_resolution)

    def verify_path(self) -> str:
        target_path = self.entry_path.get()
        path_resolution = self.path_resolution.get().lower()
        if not string_parser.valid_path(target_path, path_resolution):
            self.entry_path.state(["invalid"])
        else:
            self.entry_path.state(["!invalid"])
        return target_path

    def on_invalid_state(self, path) -> None:
        invalid_path = path.replace("Invalid path: ", "")
        self.entry_path.delete(0, tk.END)
        self.entry_path.insert(tk.END, f"Invalid path: {invalid_path}")

    def toogle_path_resolution(self, event) -> None:
        relative, absolute = "relative", "absolute"
        current_resolution = self.path_resolution.get().lower()
        if relative in current_resolution:
            self.btn_resolution["text"] = f"{absolute} path"
            self.path_resolution.set(absolute.capitalize())
        elif absolute in current_resolution:
            self.btn_resolution["text"] = f"{relative} to file"
            self.path_resolution.set(relative.capitalize())

    def _get_path_resolution(self) -> None:
        path_resolution: str = io_toml.load_toml_value(FILE_PATHS.config, "subtitle_post_processing.path_resolution")
        self.path_resolution.set(path_resolution.capitalize())


class SearchThreshold(tk.Frame):
    def __init__(self, parent) -> None:
        tk.Frame.__init__(self, parent)
        self.configure(bg=cfg.color.default_bg)
        self.pct_threashold = io_toml.load_toml_value(FILE_PATHS.config, "subtitle_filters.accept_threshold")
        self.current_value = tk.IntVar()
        self.current_value.set(self.pct_threashold)
        self.pct_value = tk.StringVar()
        self.pct_value.set(f"{self.pct_threashold} %")

        frame_text = tk.Frame(self, bg=cfg.color.default_bg)
        frame_slider = tk.Frame(self, bg=cfg.color.default_bg)

        label_description = tk.Label(frame_text, text=f"Include subtitles matching video filename by")
        label_description.configure(bg=cfg.color.default_bg, fg=cfg.color.default_fg, font=cfg.font.cas8b)
        label_description.pack(side=tk.LEFT, anchor="n")

        self.label_pct = tk.Label(frame_text, textvariable=self.pct_value, width=4)
        self.label_pct.configure(bg=cfg.color.default_bg, font=cfg.font.cas8b)
        self.label_pct.pack(side=tk.LEFT, anchor="n")

        self.set_label_color()

        self.slider = ttk.Scale(
            frame_slider,
            from_=0,
            to=100,
            orient="horizontal",
            variable=self.current_value,
            length=500,
        )
        self.slider.pack(side=tk.LEFT, anchor="n")

        frame_text.pack(side=tk.TOP, anchor="n")
        frame_slider.pack(side=tk.TOP, anchor="n")
        self.slider.bind("<Enter>", self.enter_button)
        self.slider.bind("<Leave>", self.leave_button)

    def get_value(self) -> int:
        return self.current_value.get()

    def set_value(self, event) -> None:
        self.set_label_color()
        self.pct_value.set(f"{self.get_value()} %")

    def enter_button(self, event) -> None:
        btn = event.widget
        if btn == self.slider:
            self.slider.bind("<ButtonPress>", self.press_slider)
            self.slider.bind("<Double-ButtonRelease-1>", self.dubble_press)

    def leave_button(self, event) -> None:
        btn = event.widget
        if btn == self.slider:
            self.slider.configure(command=self.set_value)

    def dubble_press(self, event) -> None:
        btn = event.widget
        if btn == self.slider:
            self.current_value.set(90)
            self.pct_value.set(f"{90} %")
            self.slider.update()
            self.update_config()
            self.set_label_color()

    def press_slider(self, event) -> None:
        btn = event.widget
        if btn == self.slider:
            self.slider.configure(command=self.set_value)
            self.slider.bind("<ButtonRelease>", self.release_slider)

    def release_slider(self, event) -> None:
        btn = event.widget
        if btn == self.slider:
            self.slider.configure(command=self.set_value)
            self.update_config()

    def update_config(self) -> None:
        value = self.current_value.get()
        io_toml.update_toml_key(FILE_PATHS.config, "subtitle_filters.accept_threshold", value)

    def set_label_color(self) -> None:
        value = self.current_value.get()
        if value in range(75, 101):
            self.label_pct.configure(fg=cfg.color.green)
        elif value in range(50, 75):
            self.label_pct.configure(fg=cfg.color.green_brown)
        elif value in range(25, 50):
            self.label_pct.configure(fg=cfg.color.red_brown)
        elif value in range(0, 25):
            self.label_pct.configure(fg=cfg.color.red)





import tkinter as tk

from subsearch.globals.constants import APP_PATHS, FILE_PATHS
from subsearch.gui.resources import config as cfg
from subsearch.utils import io_toml, io_winreg


class WindowPosition(tk.Frame):
    def __init__(self, parent) -> None:
        tk.Frame.__init__(self, parent)

    def set(self, **kwargs) -> str:
        w = kwargs.get("w", cfg.size.width)
        h = kwargs.get("h", cfg.size.height)
        ws_value_offset = kwargs.get("ws_value_offset", 0)
        hs_value_offset = kwargs.get("hs_value_offset", 0)
        ws = self.winfo_screenwidth() + ws_value_offset
        hs = self.winfo_screenheight() + hs_value_offset
        x = int((ws / 2) - (w / 2))
        y = int((hs / 2) - (h / 2))
        value = f"{w}x{h}+{x}+{y}"
        return value


class ToolTip(tk.Toplevel):
    def __init__(self, parent, widget, *text_strings, **kwargs) -> None:
        self.parent = parent
        self.widget = widget
        self.text = text_strings
        self.background = kwargs.get("bg", cfg.color.light_black)

    def show(self) -> None:
        tk.Toplevel.__init__(self, self.parent)
        self.configure(background=self.background)
        # remove the standard window titlebar from the tooltip
        self.overrideredirect(True)
        # unpack *args and put each /n on a new line
        lines = "\n".join(self.text)
        frame = tk.Frame(self, background=self.background)
        label = tk.Label(
            frame,
            text=lines,
            background=self.background,
            foreground=cfg.color.default_fg,
            justify="left",
        )
        # get size of the label to use later for positioning and sizing of the tooltip
        x, y = label.winfo_reqwidth(), label.winfo_reqheight()
        # set the size of the tooltip background to be 1px larger than the label
        frame.configure(width=x + 1, height=y + 1)

        widget_posx = self.widget.winfo_rootx()
        widget_width = self.widget.winfo_reqwidth()
        widget_center = round(widget_width / 2)
        btn_pos_middle = widget_posx + widget_center
        frame_width = frame.winfo_reqwidth()
        frame_center = round(frame_width / 2)
        center_tip_over_mouse = btn_pos_middle - frame_center
        # offset the frame 1px from edge of the tooltip corner
        frame.place(x=1, y=1, anchor="nw")
        label.place(x=0, y=0, anchor="nw")
        root_x = center_tip_over_mouse  # offset the tooltip by extra 4px so it doesn't overlap the widget
        root_y = self.widget.winfo_rooty() - self.widget.winfo_height() - 4
        # set position of the tooltip, size and add 2px around the tooltip for a 1px border
        self.geometry(f"{x+2}x{y+2}+{root_x}+{root_y}")

    def hide(self) -> None:
        self.destroy()


def configure_root(root) -> None:
    if io_toml.load_toml_value(FILE_PATHS.config, "gui.context_menu"):
        io_winreg.add_context_menu()
    root.configure(background=cfg.color.default_bg)
    root.iconbitmap(APP_PATHS.gui_assets / "subsearch.ico")
    root.geometry(WindowPosition.set(root))  # type: ignore
    root.resizable(False, False)



class Error(Exception):
    pass


class ProviderError(Error):
    pass


class CaptchaError(ProviderError):
    def __init__(self, message: str = "Captcha challenge encountered") -> None:
        super().__init__(message)


class ProviderNotImplemented(ProviderError):
    def __init__(self, message: str = "This provider is not implemented") -> None:
        super().__init__(message)


class MultipleInstancesError(Error):
    def __init__(self, mutex_name: str) -> None:
        error_type = f"Multiple instances of the application are running."
        error_msg = f"Unable to acquire mutex '{mutex_name}' because it's already locked. Please close any other instances of the application before launching a new one."
        additional_info = "If you need to run multiple instances, you can adjust your app settings to allow this, but be aware that it's not recommended'."
        message = f"{error_type}\n{error_msg}\n{additional_info}"
        super().__init__(message)



import threading
from typing import Any, Callable

from subsearch import core
from subsearch.globals import log


class CreateThread(threading.Thread):
    def __init__(self, *args, **kwargs) -> None:
        self._target = None
        self._args = ()
        self._kwargs = {}  # type: ignore
        super().__init__(*args, **kwargs)
        self.exception = None
        self.return_value = None
        assert self._target

    def run(self) -> None:
        try:
            log.stdout(f"Thread {self.name} started", level="debug", print_allowed=False)
            if self._target:
                self.return_value = self._target(*self._args, **self._kwargs)
        except Exception as e:
            self.exception = e  # type: ignore
        finally:
            del self._target, self._args, self._kwargs

    def join(self, timeout=None) -> Any:
        super().join(timeout)
        if self.exception:
            msg = f"\nError occurred in thread {self.name}\n\tTraceback (most recent call last):\n\t\t{self.exception}"
            log.stdout(msg, level="error", print_allowed=False)
            raise self.exception
        else:
            log.stdout(f"Thread {self.name} finished", level="debug", print_allowed=False)
        return self.return_value


def _create_threads(*tasks) -> None:
    threads = []
    for target in tasks:
        name = f"{target.__name__.lower()}"
        task_thread = CreateThread(target=target, name=name)
        task_thread.start()
        threads.append(task_thread)

    for t in threads:
        t.join()


def _start_search(cls: "core.SubsearchCore", provider: Callable[..., Any], flag: str) -> None:
    search_provider = provider(**cls.search_kwargs)
    search_provider.start_search(flag=flag)
    cls.accepted_subtitles.extend(search_provider.accepted_subtitles)
    cls.rejected_subtitles.extend(search_provider.rejected_subtitles)



import threading

import pystray
from PIL import Image

from subsearch.globals import decorators, log, metaclasses
from subsearch.globals.constants import APP_PATHS, VERSION


class SystemTray(metaclass=metaclasses.Singleton):
    @decorators.system_tray_conditions
    def __init__(self) -> None:
        title = f"Subsearch {VERSION}"
        icon = Image.open(APP_PATHS.gui_assets / "subsearch.ico")
        self.tray = pystray.Icon("Subsearch", icon=icon, title=title)
        self.thread_tray = threading.Thread(daemon=True, target=self._run_pystray)

    @decorators.system_tray_conditions
    def display_toast(self, title: str, msg: str) -> None:
        self.tray.title
        self.tray.notify(msg, title)

    @decorators.system_tray_conditions
    def start(self) -> None:
        log.stdout("Subsearch was added to the system tray", level="debug")
        self.thread_tray.start()

    @decorators.system_tray_conditions
    def stop(self) -> None:
        self.tray.stop()
        self.thread_tray.join()
        log.stdout("Subsearch was removed from the system tray", level="debug")

    def _run_pystray(self) -> None:
        self.tray.run()



from typing import no_type_check

from subsearch.providers import common_utils


class AdvancedSearch:
    def __init__(self, *args, **kwargs) -> None: ...

    def get_url(self) -> str: ...


class ImdbMovieSearch(AdvancedSearch):
    def __init__(self, title: str, year: int, *args, **kwargs) -> None:
        super().__init__(self, *args, **kwargs)
        self.title = title
        self.year = year

    def get_url(self) -> str:
        imdb_domain = "https://www.imdb.com"
        return f"{imdb_domain}/search/title/?{self._title_search}&{self._release_date_search}"

    @property
    def _prior_year(self) -> str:
        return f"{self.year-1}-01-01"

    @property
    def _release_year(self) -> str:
        return f"{self.year}-12-31"

    @property
    def _release_title(self) -> str:
        return self.title.replace(" ", "+")

    @property
    def _title_search(self) -> str:
        return f"title={self._release_title}"

    @property
    def _release_date_search(self) -> str:
        return f"release_date={self._prior_year},{self._release_year}"


class ImdbTvseriesSearch(AdvancedSearch):
    def __init__(self, title: str, *args, **kwargs) -> None:
        super().__init__(self, *args, **kwargs)
        self.title = title

    def get_url(self) -> str:
        imdb_domain = "https://www.imdb.com"
        return f"{imdb_domain}/search/title/?{self._title_search}"

    @property
    def _release_title(self) -> str:
        return self.title.replace(" ", "+")

    @property
    def _title_search(self) -> str:
        return f"title={self._release_title}"


class FindImdbID:
    @no_type_check
    def __init__(self, title: str, year: int, tvseries: bool, request_timeout=(4, 8)) -> None:
        self.title = title.lower()
        self.year = year
        self.tvseries = tvseries
        self.imdb_id = ""
        adv_search = self._advanced_search()

        url = adv_search.get_url()
        tree = common_utils.request_parsed_response(url=url, timeout=request_timeout)
        if not tree:
            return None

        product = tree.css("a.ipc-title-link-wrapper h3.ipc-title__text")

        for item in product:
            href_ = item.parent.attrs["href"]
            imdb_id = href_.split("/")[2]
            title_ = item.text().split(". ")[-1]
            _year = item.parent.parent.next.child.child.html
            release_year = self._handle_ongoing_show(_year)

            if self.title != title_.lower():
                continue
            if not self.tvseries:
                if self.year != release_year and (self.year - 1) != release_year:
                    continue

            self.imdb_id = str(imdb_id)
            break

    def _advanced_search(self) -> AdvancedSearch:
        if self.tvseries:
            return ImdbTvseriesSearch(self.title)
        else:
            return ImdbMovieSearch(self.title, self.year)

    def _handle_ongoing_show(self, year: str) -> int:
        if not self.tvseries:
            release_year = year.split("–")[0]
            return int(release_year)



from selectolax.parser import Node

from subsearch.providers import common_utils


class YifySubtitlesScraper(common_utils.ProviderHelper):
    def __init__(self, *args, **kwargs) -> None:
        common_utils.ProviderHelper.__init__(self, *args, **kwargs)
        self.provider_name = ""

    def get_subtitles(self) -> None:
        tree = common_utils.request_parsed_response(url=self.url_yifysubtitles, timeout=self.request_timeout)
        if not tree:
            return None
        
        product = tree.select("tr")
        for item in product.matches[1:]:  # type: ignore
            if self.skip_item(item):
                continue
            self.parse_item(item)

    def parse_item(self, item) -> None:
        node = item.css_first("a")
        titles = node.text().strip().split("subtitle ")[-1].split("\n")
        _href = node.attributes["href"].split("/")  # type: ignore
        href = _href[-1]
        download_url = f"https://yifysubtitles.org/subtitle/{href}.zip"
        for subtitle_name in titles:
            self.prepare_subtitle(self.provider_name, subtitle_name, download_url, {})

    def skip_item(self, item: Node) -> bool:
        subtitle_language = item.css_first("span.sub-lang").child.text_content  # type: ignore
        subtitle_hi = item.css_matches("span.hi-subtitle")
        if subtitle_language.lower() != self.current_language.lower():  # type: ignore
            return True
        if not self.subtitle_hi_match(subtitle_hi):
            return True
        return False


class YifiSubtitles(YifySubtitlesScraper):
    def __init__(self, *args, **kwargs) -> None:
        YifySubtitlesScraper.__init__(self, *args, **kwargs)
        self.provider_name = self.__class__.__name__.lower()

    def start_search(self, *args, **kwargs) -> None:
        self.get_subtitles()



import tkinter as tk
from tkinter import IntVar, ttk

from subsearch.globals.constants import FILE_PATHS
from subsearch.gui import common_utils
from subsearch.utils import io_toml


class SelectLanguage(ttk.Labelframe):
    def __init__(self, parent) -> None:
        ttk.Labelframe.__init__(self, parent)
        self.configure(text="Languages", padding=10)
        self.active_btn = None
        self.rownum = 0
        self.colnum = 1
        self.checkbox_values = {}
        self.name_find_key = {}
        self.tip_present = False
        self.languages = io_toml.load_toml_data(FILE_PATHS.language_data)
        self.current_language = io_toml.load_toml_value(FILE_PATHS.config, "subtitle_filters.current_language")
        self.checkbuttons: dict[ttk.Checkbutton, IntVar] = {}
        frame = None
        for enum, (language, language_data) in enumerate(self.languages.items()):
            if enum % 14 == 0:
                frame = ttk.Frame(self)
                frame.pack(side=tk.LEFT, anchor="n")

            valuevar = tk.IntVar()
            btn = ttk.Checkbutton(frame, text=language_data["name"], onvalue=1, offvalue=9, variable=valuevar, width=20)
            if self.current_language == language:
                valuevar.set(1)
                self.checkbox_values[btn] = valuevar
            else:
                valuevar.set(0)
                self.checkbox_values[btn] = valuevar
            btn.pack(padx=2, pady=9)
            if valuevar.get() == 1:
                self.active_btn = btn

            btn.bind("<Enter>", self.enter_button)
            btn.bind("<Leave>", self.leave_button)
            self.name_find_key[language_data["name"]] = language

    def enter_button(self, event) -> None:
        btn = event.widget
        toml_key = self.name_find_key[btn["text"]]
        providers = ", ".join(([_i.title() for _i in self.languages[toml_key]["incompatibility"]]))
        if providers:
            tip_text = f"If enabled, '{providers}' will be automatically skipped."
            self.tip = common_utils.ToolTip(btn, btn, tip_text)
            self.tip.show()
            self.tip_present = True
        btn.bind("<ButtonPress-1>", self.press_button)

    def leave_button(self, event) -> None:
        btn = event.widget
        if self.tip_present:
            self.tip.hide()
            self.tip_present = False

    def press_button(self, event) -> None:
        btn = event.widget
        btn.bind("<ButtonRelease-1>", self.set_current_language)

    def set_current_language(self, event) -> None:
        btn = event.widget
        self.checkbox_values[self.active_btn].set(0)  # type: ignore
        self.active_btn.state(["!alternate"])  # type: ignore
        self.active_btn = btn
        value = self.name_find_key[btn["text"]]
        io_toml.update_toml_key(FILE_PATHS.config, "subtitle_filters.current_language", value)



import dataclasses
import inspect
import threading
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Optional

import logging

from subsearch.globals import metaclasses
from subsearch.globals.constants import APP_PATHS, FILE_PATHS
from subsearch.globals.dataclasses import GenericDataClass


def capture_call_info(func: Callable[..., Any]) -> Callable[..., Any]:
    def wrapper(*args, **kwargs) -> Any:
        frame = inspect.currentframe().f_back  # type: ignore
        current_time = datetime.now().time()
        call_time = current_time.strftime("%H:%M:%S.%f")[:-3]
        kwargs["call_module"] = frame.f_globals["__name__"].split(".")[-1]  # type: ignore
        kwargs["call_lineno"] = frame.f_lineno  # type: ignore
        kwargs["call_ct"] = call_time
        return func(*args, **kwargs)

    return wrapper


class ANSIEscapeSequences:
    """
    ANSI escape sequences for text colors
    """

    RESET = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


class _Logging(metaclass=metaclasses.Singleton):
    def __init__(self, *args, **kwargs) -> None:
        self.logger_name = kwargs.get("logger_name", "subsearch")
        if not APP_PATHS.appdata_subsearch.exists():
            APP_PATHS.appdata_subsearch.mkdir(parents=True, exist_ok=True)
        self.log_file_path = kwargs.get("debug_log_file", FILE_PATHS.log)
        self._debug_logger = self.create_logger()
        self._lock = threading.Lock()
        self._ansi = ANSIEscapeSequences

    def create_logger(self) -> logging.Logger:
        logger = logging.getLogger(self.logger_name)
        logger.setLevel(logging.DEBUG)
        file_handler = logging.FileHandler(self.log_file_path, mode="w")
        file_handler.setLevel(logging.DEBUG)
        formatter = logging.Formatter("%(message)s", datefmt="%H:%M:%S")
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
        return logger

    def log(self, message: str, **kwargs) -> None:
        with self._lock:
            self._log(message, **kwargs)
            self._print(message, **kwargs)

    def _log(self, message: str, **kwargs) -> None:
        level = kwargs.get("level", "info")
        log_methods = {
            "debug": self._debug_logger.debug,
            "info": self._debug_logger.info,
            "warning": self._debug_logger.warning,
            "error": self._debug_logger.error,
            "critical": self._debug_logger.critical,
        }
        module = kwargs.get("call_module")
        lineno = kwargs.get("call_lineno")
        time = kwargs.get("call_ct")
        log_msg = f"{level.upper()} {module}:{lineno} {time}: {message}"
        log_methods[level](log_msg)

    def _print(self, message, **kwargs) -> None:
        print_allowed = kwargs.get("print_allowed", True)
        hex_color = kwargs.get("hex_color", "#")
        style = kwargs.get("style", "")

        if not print_allowed:
            pass
        elif hex_color == "#" and len(hex_color) == 1:
            print(message)
        elif hex_color.startswith("#") and len(hex_color[1:]) == 6:
            self._color_print(message, hex_color, style)

    def _hex_to_ansi(self, hex_color: str) -> str:
        r, g, b = int(hex_color[1:3], 16), int(hex_color[3:5], 16), int(hex_color[5:7], 16)
        return f"\033[38;2;{r};{g};{b}m"

    def _color_print(self, message: str, hex_color: str, style: str) -> None:
        color_code = self._hex_to_ansi(hex_color)
        style_code = getattr(self._ansi, style.upper()) if style != "" else ""
        print(f"{style_code}{color_code}{message}{self._ansi.RESET}")


class Logger(metaclass=metaclasses.Singleton):
    def __init__(self) -> None:
        self._logger = _Logging()

    def __call__(self, message: str, **kwargs) -> None:
        self.log(message, **kwargs)

    def log(self, message: str, **kwargs) -> None:
        end_new_line = kwargs.get("end_new_line", False)
        self._logger.log(message, **kwargs)
        if end_new_line:
            self._logger.log("", **kwargs)

    @capture_call_info
    def stdout(self, message: str, **kwargs) -> None:
        self(message, **kwargs)

    @capture_call_info
    def brackets(self, message: str, **kwargs) -> None:
        self(f"--- [{message}] ---", hex_color="#fab387", style="bold", **kwargs)

    @capture_call_info
    def subtitle_match(self, provider: str, subtitle_name: str, result: int, threshold: int, **kwargs) -> None:
        if result >= threshold:
            self(f"{provider:<14}{result:>3}% {subtitle_name}", hex_color="#a6e3a1", **kwargs)
        else:
            self(f"{provider:<14}{result:>3}% {subtitle_name}", **kwargs)

    @capture_call_info
    def file_system_action(self, action_type: str, src: Path, dst: Optional[Path] = None, **kwargs) -> None:
        if src.is_file():
            type_ = "file"
        elif src.is_dir():
            type_ = "directory"
        else:
            type_ = "item"

        __src = src.relative_to(src.parent.parent) if src else None
        __dst = dst.relative_to(dst.parent.parent) if dst else None

        action_messages = {
            "remove": rf"Removing {type_}: ...\{__src}",
            "rename": rf"Renaming {type_}: ...\{__src} -> ...\{__dst}",
            "move": rf"Moving {type_}: ...\{__src} -> ...\{__dst}",
            "extract": rf"Extracting archive: ...\{__src} -> ...\{__dst}",
        }

        message = action_messages.get(action_type)

        if not message:
            raise ValueError("Invalid action type")

        self(message, **kwargs)

    @capture_call_info
    def dataclass(self, instance: GenericDataClass, **kwargs) -> None:
        if not dataclasses.is_dataclass(instance):
            raise ValueError("Input is not a dataclass instance.")
        instance_name = f"--- [{instance.__class__.__name__}] ---"
        self(instance_name, hex_color="#fab387", style="bold", **kwargs)
        for field in dataclasses.fields(instance):
            key = field.name
            value = getattr(instance, key)
            padding = " " * (30 - len(key))
            self(f"{key}:{padding}{value}", **kwargs)

    @capture_call_info
    def task_completed(self, **kwargs) -> None:
        self("Tasks completed", level="info", hex_color="#89b4fa", **kwargs)



import re
from typing import Any

from subsearch.globals import log
from subsearch.providers import common_utils


class OpenSubtitlesScraper(common_utils.ProviderHelper):
    def __init__(self, *args, **kwargs) -> None:
        common_utils.ProviderHelper.__init__(self, *args, **kwargs)
        self.provider_name = ""

    def is_opensubtitles_down(self, tree: Any) -> bool:
        is_offline = tree.css_matches("pre")
        if is_offline is False:
            return False
        offline_text = tree.css_first("pre").text()
        if offline_text.startswith("Site will be online soon"):
            log.stdout(f"opensubtitles is down: {offline_text}", level="error")
            return True
        return False

    def get_subtitles(self, url: str) -> None:
        tree = common_utils.request_parsed_response(url=url, timeout=self.request_timeout)
        if not tree:
            return None
        items = tree.css("item")
        if self.is_opensubtitles_down(tree):
            return None
        for item in items:
            download_url = item.css_first("enclosure").attributes["url"]
            released_as = item.css_first("description").child.text_content.strip()  # type: ignore
            subtitle_name = re.findall("^.*?: (.*?);", released_as)[0]  # https://regex101.com/r/LWAmJK/1
            self.prepare_subtitle(self.provider_name, subtitle_name, download_url, {})

    def with_hash(self, url: str, subtitle_name: str) -> None:
        tree = common_utils.request_parsed_response(url=url, timeout=self.request_timeout)
        if not tree:
            return None
        if self.is_opensubtitles_down(tree):
            return None
        try:
            sub_id = tree.css_first("#bt-dwl-bt").attributes["data-product-id"]
        except AttributeError:
            return None
        download_url = f"https://dl.opensubtitles.org/en/download/sub/{sub_id}"
        self.prepare_subtitle(self.provider_name, subtitle_name, download_url)


class OpenSubtitles(OpenSubtitlesScraper):
    def __init__(self, *args, **kwargs) -> None:
        OpenSubtitlesScraper.__init__(self, *args, **kwargs)
        self.provider_name = self.__class__.__name__.lower()

    def start_search(self, flag: str) -> None:
        flags = {"hash": self.search_site, "site": self.search_hash}
        flags[flag]()

    def search_hash(self) -> None:
        self.with_hash(self.url_opensubtitles_hash, self.release)

    def search_site(self) -> None:
        self.get_subtitles(self.url_opensubtitles)



from dataclasses import dataclass


@dataclass(order=True, frozen=True)
class Size:
    width: int = 800
    height: int = 700
    max_content_width = width - 20


@dataclass(order=True, frozen=True)
class Position:
    screen_x: int = int(Size().width / 2)
    screen_y: int = int(Size().height / 2 - 41)
    screen_hidden_x: int = int(Size().width * 2)


@dataclass(order=True, frozen=True)
class Color:
    purple: str = "#b294bb"
    red: str = "#bc473b"
    dark_red: str = "#89332a"
    black_red: str = "#4b1713"
    red_brown: str = "#b26947"
    orange: str = "#ab7149"
    light_orange: str = "#de935f"
    yellow: str = "#f0c674"
    blue: str = "#81a2be"
    cyan: str = "#82b3ac"
    green: str = "#9fa65d"
    green_brown: str = "#a59256"
    grey: str = "#4c4c4c"
    light_grey: str = "#727272"
    silver_grey: str = "#8a8a8a"
    default_fg: str = "#bdbdbd"
    default_bg: str = "#1a1b1b"
    default_bg_dark: str = "#111111"
    light_black: str = "#0e0e0e"
    black: str = "#151515"
    dark_black: str = "#000000"


@dataclass(order=True, frozen=True)
class Font:
    cas6b: str = "Cascadia 6 bold"
    cas8: str = "Cascadia 8"
    cas8i: str = "Cascadia 8 italic"
    cas8b: str = "Cascadia 8 bold"
    cas10b: str = "Cascadia 10 bold"
    cas11: str = "Cascadia 11"
    cas20b: str = "Cascadia 20 bold"


size = Size()
position = Position()
font = Font()
color = Color()


DEFAULT_LABEL_CONFIG = dict(bg=color.default_bg, fg=color.default_fg, font=font.cas8b)
DEFAULT_LABEL_GRID = dict(row=0, column=0, sticky="w", padx=2, pady=2)
DEFAULT_BTN_TOGGLE_GRID = dict(row=0, column=2, pady=2)



import ctypes
import functools
import sys
from typing import Any, Callable

from subsearch.globals import exceptions
from subsearch.globals.constants import FILE_PATHS, GUID
from subsearch.utils import io_toml

enable_system_tray: bool


def apply_mutex(func: Callable) -> Callable:
    def inner(*args, **kwargs) -> Any:
        try:
            if not io_toml.load_toml_value(FILE_PATHS.config, "misc.single_instance"):
                return func()
        except FileNotFoundError:
            pass
        except KeyError:
            pass
        except TypeError:
            pass
        kernel32 = ctypes.WinDLL("kernel32")
        mutex = kernel32.CreateMutexW(None, False, GUID)
        last_error = kernel32.GetLastError()

        if last_error == 183:
            raise exceptions.MultipleInstancesError(GUID)
        try:
            kernel32.WaitForSingleObject(mutex, -1)
            return func(*args, **kwargs)
        finally:
            kernel32.ReleaseMutex(mutex)

    return inner


def check_option_disabled(func) -> Callable[..., Any]:
    def wrapper(self, event, *args, **kwargs) -> Any:
        btn = event.widget
        if btn.instate(["disabled"]):
            return None
        return func(self, event, *args, **kwargs)

    return wrapper


def call_func(func) -> Callable[..., Any]:

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        func_name = f"{func.__name__}"
        if not args[0].call_conditions.call_func(func_name=func_name, *args, **kwargs):
            return None
        return func(*args, **kwargs)

    return wrapper


def system_tray_conditions(func) -> Callable[..., Any]:
    def wrapper(*args, **kwargs) -> Any:
        if enable_system_tray:
            return func(*args, **kwargs)

    return wrapper


def except_hook(func, excepthook_) -> Callable[..., Any]:
    def wrapper(*args, **kwargs) -> Any:
        sys.excepthook = excepthook_
        return func(*args, **kwargs)

    return wrapper



import tkinter as tk
from typing import Any

from subsearch.globals import log
from subsearch.gui import common_utils, resource_loader, root
from subsearch.gui.resources import config as cfg
from subsearch.gui.screens import (
    download_manager,
    language_options,
    search_options,
    subsearch_options,
)


class ScreenManager(tk.Frame):
    def __init__(self, parent, available_screens: dict[str, Any], active_screen: str) -> None:
        tk.Frame.__init__(self, parent)
        self.configure(bg=cfg.color.default_bg_dark, width=cfg.size.width, height=82)
        relx_value = 0.0
        btn_kwargs: dict[str, Any] = dict(
            master=self, width=54, height=54, bg=cfg.color.default_bg_dark, highlightthickness=0
        )
        self.parent = parent
        self.available_screens = available_screens
        self.buttons = {}

        for tab_key in available_screens.keys():
            self.buttons[tab_key] = tk.Canvas(**btn_kwargs)

        for btn_key, btn_widget in self.buttons.items():
            relx_value += 0.2
            btn_widget.place(relx=relx_value, rely=0.5, anchor="center")
            btn_widget.bind("<Enter>", self.enter_menu_btn)
            btn_widget.bind("<Leave>", self.leave_menu_btn)
            resource_loader.asset_menu_btn(btn_widget, btn_key, "rest")

        self.active_screen = active_screen
        self.activate_screen()

    def activate_screen(self) -> None:
        self.available_screens[self.active_screen].place(
            x=cfg.position.screen_x, y=cfg.position.screen_y, anchor="center"
        )
        resource_loader.asset_menu_btn(self.buttons[self.active_screen], self.active_screen, "press")
        title_tab = self.active_screen.capitalize().replace("_", " ")
        self.parent.title(f"Subsearch - {title_tab}")

    def deactivate_screen(self) -> None:
        for btn_key, btn_widget in self.buttons.items():
            if self.active_screen == btn_key:
                continue
            self.available_screens[btn_key].place(x=cfg.position.screen_hidden_x, y=cfg.position.screen_y, anchor="nw")
            resource_loader.asset_menu_btn(btn_widget, btn_key, "rest")

    def enter_menu_btn(self, event) -> None:
        _btn_key, btn_widget = self.get_btn(self.buttons, event)
        btn_widget.bind("<ButtonPress>", self.press_menu_btn)
        resource_loader.asset_menu_btn(self.buttons[_btn_key], _btn_key, "hover", y=23)

    def leave_menu_btn(self, event) -> None:
        _btn_key, btn_value = self.get_btn(self.buttons, event)
        btn_value.unbind("<ButtonPress>")
        resource_loader.asset_menu_btn(
            self.buttons[_btn_key], _btn_key, "rest" if _btn_key != self.active_screen else "press"
        )

    def press_menu_btn(self, event) -> None:
        btn_key, btn_widget = self.get_btn(self.buttons, event, False)
        btn_widget.bind("<ButtonRelease>", self.release_menu_btn)
        resource_loader.asset_menu_btn(btn_widget, btn_key, "press", y=20)

    def release_menu_btn(self, event) -> None:
        btn_key, _btn_widget = self.get_btn(self.buttons, event)
        self.active_screen = btn_key
        self.activate_screen()
        self.deactivate_screen()

    def get_btn(self, btns: dict, event_: Any, equals=True) -> tuple | None:
        for btn_key, btn_widget in btns.items():
            if event_.widget == btn_widget and equals:
                return btn_key, btn_widget
            if event_.widget == btn_widget and not equals:
                return btn_key, btn_widget


class LanguageOptions(tk.Frame):
    def __init__(self, parent) -> None:
        tk.Frame.__init__(self, parent, width=cfg.size.width, height=cfg.size.height)
        self.configure(bg=cfg.color.default_bg)
        language_options.SelectLanguage(self).pack(anchor="center", expand=True)


class SearchOptions(tk.Frame):
    def __init__(self, parent) -> None:
        tk.Frame.__init__(self, parent, width=cfg.size.width, height=cfg.size.height)
        self.configure(bg=cfg.color.default_bg)
        group_a = tk.Frame(self, bg=cfg.color.default_bg)
        group_a.pack(anchor="center", expand=True, fill="both")
        search_options.SubtitleFilters(group_a).pack(side=tk.LEFT, anchor="center", expand=True, fill="both", padx=2)
        search_options.Providers(group_a).pack(side=tk.LEFT, anchor="center", expand=True, fill="both", padx=2)
        tk.Frame(self, height=30, bg=cfg.color.default_bg).pack(anchor="center", expand=True)
        search_options.SubtitlePostProcessing(self).pack(anchor="center", expand=True, fill="x")
        search_options.SubtitlePostProcessingDirectory(self).pack(anchor="center", expand=True, fill="x")
        tk.Frame(self, height=60, bg=cfg.color.default_bg).pack(anchor="center", expand=True)
        search_options.SearchThreshold(self).pack(anchor="center")


class SubsearchOptions(tk.Frame):
    def __init__(self, parent) -> None:
        tk.Frame.__init__(self, parent, width=cfg.size.width, height=cfg.size.height)
        self.configure(bg=cfg.color.default_bg)
        subsearch_options.FileExtensions(self).pack(anchor="center", fill="x")
        tk.Frame(self, height=15, bg=cfg.color.default_bg).pack(anchor="center", expand=True)
        subsearch_options.SubsearchOption(self).pack(anchor="center", fill="x")
        tk.Frame(self, height=15, bg=cfg.color.default_bg).pack(anchor="center", expand=True)
        subsearch_options.DownloadManagerOptions(self).pack(anchor="center", fill="x")
        tk.Frame(self, height=15, bg=cfg.color.default_bg).pack(anchor="center", expand=True)
        subsearch_options.AdvancedUser(self).pack(anchor="center", fill="x")
        tk.Frame(self, height=15, bg=cfg.color.default_bg).pack(anchor="center", expand=True)
        subsearch_options.CheckForUpdates(self)


class DownloadManager(tk.Frame):
    def __init__(self, parent, **kwargs) -> None:
        tk.Frame.__init__(self, parent, width=cfg.size.width, height=cfg.size.height)
        self.configure(bg=cfg.color.default_bg)
        download_manager.DownloadManager(self, **kwargs).pack(anchor="center", expand=True, fill="x")
        enforce_width = tk.Frame(self, height=0, width=cfg.size.max_content_width, bg=cfg.color.default_bg)
        enforce_width.pack(anchor="center", expand=True)


def open_screen(tab_name: str, **kwargs) -> None:
    root.bind("<KeyPress>", close_mainloop)
    common_utils.configure_root(root)
    resource_loader.set_ttk_theme(root)
    screens = {
        "language_options": LanguageOptions(root),
        "search_options": SearchOptions(root),
        "subsearch_options": SubsearchOptions(root),
        "download_manager": DownloadManager(root, **kwargs),
    }
    manager = ScreenManager(root, screens, tab_name.lower())
    manager.place(y=cfg.size.height - 82)
    log.stdout("GUI is running", level="debug")
    root.mainloop()


def close_mainloop(event) -> None:
    if event.keysym == "Escape":
        root.destroy()



from subsearch.utils import io_app

DEVICE_INFO = io_app.get_system_info()
VIDEO_FILE = io_app.get_video_file_data()
APP_PATHS = io_app.get_app_paths()
FILE_PATHS = io_app.get_file_paths()
SUPPORTED_FILE_EXT = io_app.get_supported_file_ext()
SUPPORTED_PROVIDERS = io_app.get_supported_providers()
DEFAULT_CONFIG = io_app.get_default_app_config()
REGISTRY_PATHS = io_app.get_windows_registry_paths()
COMPUTER_NAME = io_app.get_computer_name()
VERSION = io_app.get_app_version()
GUID = io_app.get_guid()



import sys
import winreg
from pathlib import Path

from subsearch.globals import log
from subsearch.globals.constants import (
    APP_PATHS,
    COMPUTER_NAME,
    DEVICE_INFO,
    FILE_PATHS,
    REGISTRY_PATHS,
)
from subsearch.utils import io_toml


class LaunchOptions:
    def __init__(self) -> None:
        self.show_terminal = io_toml.load_toml_value(FILE_PATHS.config, "gui.show_terminal")
        self.python_path = Path(sys.executable).parent
        self.console_title = "import ctypes; ctypes.windll.kernel32.SetConsoleTitleW('subsearch');"
        self.working_directory = f"import os; os.chdir(r'{APP_PATHS.home}');"
        self.import_subsearch = "import subsearch; subsearch.main()"

    def _get_mode_executable(self) -> str:
        return f'"{sys.argv[0]}" "%1"'

    def _get_mode_interpreter(self) -> str:
        python_executable = {
            True: f'{self.python_path}\\python.exe -c "{self.console_title} {self.working_directory} {self.import_subsearch}" "%1"',
            False: f'{self.python_path}\\pythonw.exe -c "{self.console_title} {self.working_directory} {self.import_subsearch}" "%1"',
        }
        return python_executable[self.show_terminal]

    def get_parameter(self) -> str:
        if DEVICE_INFO.mode == "executable":
            value = self._get_mode_executable()
        elif DEVICE_INFO.mode == "interpreter":
            value = self._get_mode_interpreter()
        return value


def get_command_value() -> str:
    launch_options = LaunchOptions()
    return launch_options.get_parameter()


def get_icon_value() -> str:
    show_icon: str = io_toml.load_toml_value(FILE_PATHS.config, "gui.context_menu_icon")
    if show_icon:
        return str(APP_PATHS.gui_assets / "subsearch.ico")
    else:
        return ""


def get_appliesto_value() -> str:
    file_ext = io_toml.load_toml_value(FILE_PATHS.config, "file_extensions")
    value = ""
    for ext, v in zip(file_ext.keys(), file_ext.values()):
        if v is True:
            value += "".join(f'".{ext}" OR ')
    if value.endswith(" OR "):  # remove last OR
        value = value[:-4]

    return value


def del_registry_key(reg_path: str, key: str) -> None:
    with winreg.ConnectRegistry(COMPUTER_NAME, winreg.HKEY_CURRENT_USER) as hkey:
        with winreg.OpenKey(hkey, reg_path, 0, winreg.KEY_WRITE) as sk:
            winreg.DeleteKey(sk, key)


def del_context_menu() -> None:
    del_registry_key(REGISTRY_PATHS.subsearch, "command")
    del_registry_key(REGISTRY_PATHS.shell, "Subsearch")


def write_keys() -> None:
    registry_keys = [
        (REGISTRY_PATHS.classes, "*"),
        (REGISTRY_PATHS.asterisk, "shell"),
        (REGISTRY_PATHS.shell, "Subsearch"),
        (REGISTRY_PATHS.subsearch, "command"),
    ]
    with winreg.ConnectRegistry(COMPUTER_NAME, winreg.HKEY_CURRENT_USER) as hkey:
        for key, sub_key in registry_keys:
            with winreg.OpenKey(hkey, key, 0, winreg.KEY_WRITE) as sk:
                winreg.CreateKey(sk, sub_key)


def set_write_valuex(sub_key: str, value_name: str, value: str) -> None:
    try:
        with winreg.ConnectRegistry(COMPUTER_NAME, winreg.HKEY_CURRENT_USER) as hkey:
            with winreg.OpenKey(hkey, sub_key, 0, winreg.KEY_WRITE) as sk:
                winreg.SetValueEx(sk, value_name, 0, winreg.REG_SZ, value)
    except FileNotFoundError:
        pass


def write_valuex(key: str) -> None:
    key_map = {
        "subsearch": {"key_type": REGISTRY_PATHS.subsearch, "value_name": "", "value": "Subsearch"},
        "icon": {"key_type": REGISTRY_PATHS.subsearch, "value_name": "Icon", "value": get_icon_value()},
        "appliesto": {"key_type": REGISTRY_PATHS.subsearch, "value_name": "AppliesTo", "value": get_appliesto_value()},
        "command": {"key_type": REGISTRY_PATHS.subsearch_command, "value_name": "", "value": get_command_value()},
    }

    key = key.lower()
    if key in key_map:
        key_type = key_map[key]["key_type"]
        value_name = key_map[key]["value_name"]
        value = key_map[key]["value"]
        set_write_valuex(key_type, value_name, value)


def write_all_valuex() -> None:
    items = ["subsearch", "icon", "appliesto", "command"]
    for i in items:
        write_valuex(i)


def add_context_menu() -> None:
    write_keys()
    write_all_valuex()


def check_long_paths_enabled() -> bool:
    value_name = "LongPathsEnabled"

    try:
        with winreg.ConnectRegistry(COMPUTER_NAME, winreg.HKEY_LOCAL_MACHINE) as hkey:
            with winreg.OpenKey(hkey, REGISTRY_PATHS.long_paths, 0, winreg.KEY_READ) as sk:
                value, _ = winreg.QueryValueEx(sk, value_name)
                return bool(value)
    except FileNotFoundError:
        return False
    except Exception as e:
        log.error(f"Failed to check long path status: {e}")
        return False



import sys
import time
from pathlib import Path

from subsearch import core
from subsearch.globals import decorators

PREF_COUNTER = time.perf_counter()
PACKAGE_PATH = Path(__file__).resolve().parent.as_posix()
HOME_PATH = Path(PACKAGE_PATH).parent.as_posix()
sys.path.append(HOME_PATH)
sys.path.append(PACKAGE_PATH)


class Subsearch:
    def __init__(self) -> None:
        self.subsearch_core = core.SubsearchCore(PREF_COUNTER)

    def start_app(self) -> None:
        self.subsearch_core.init_search(
            self.subsearch_core.opensubtitles,
            self.subsearch_core.yifysubtitles,
            self.subsearch_core.subsource,
        )
        self.subsearch_core.download_files()
        self.subsearch_core.download_manager()
        self.subsearch_core.subtitle_post_processing()
        self.subsearch_core.clean_up()

    def exit_app(self) -> None:
        self.subsearch_core.core_on_exit()


@decorators.apply_mutex
def main() -> None:
    subsearch = Subsearch()
    subsearch.start_app()
    subsearch.exit_app()


if __name__ == "__main__":
    main()



import tkinter as tk
import webbrowser
from tkinter import BooleanVar, ttk
from typing import Any

from subsearch.globals import decorators
from subsearch.globals.constants import DEVICE_INFO, FILE_PATHS, VERSION
from subsearch.gui import common_utils
from subsearch.gui.resources import config as cfg
from subsearch.utils import io_toml, io_winreg, string_parser, update


def _handle_file_extensions_check_btn(cls, parent_key) -> None:
    if "context_menu" not in parent_key:
        return None
    [instance.update_state() for instance in FileExtensions.instances]


def _handle_other_check_btn(cls, value, child_key) -> None:
    for check_button, tuple_value in cls.checkbuttons.items():
        key = tuple_value[0]
        if key != child_key:
            continue
        elif value.get():
            check_button.state(["disabled"])
        elif not value.get():
            check_button.state(["!disabled"])


class FileExtensions(ttk.Labelframe):
    instances: list["FileExtensions"] = []

    def __init__(self, parent) -> None:
        ttk.Labelframe.__init__(self, parent)
        self.configure(text="File Exstensions", padding=10)
        FileExtensions.instances.append(self)
        self.data = io_toml.load_toml_data(FILE_PATHS.config)
        self.file_extensions = io_toml.load_toml_value(FILE_PATHS.config, "file_extensions")

        self.checkbuttons: dict[ttk.Checkbutton, tuple[str, BooleanVar]] = {}

        frame = None
        for enum, (key, value) in enumerate(self.file_extensions.items()):
            if enum % 4 == 0:
                frame = ttk.Frame(self)
                frame.pack(side=tk.LEFT, anchor="n")

            boolean = tk.BooleanVar()
            boolean.set(value)
            check_btn = ttk.Checkbutton(frame, text=key, onvalue=True, offvalue=False, variable=boolean)
            if not io_toml.load_toml_value(FILE_PATHS.config, "gui.context_menu"):
                check_btn.state(["disabled"])

            check_btn.pack(padx=4, pady=4, ipadx=10)
            self.checkbuttons[check_btn] = key, boolean
            check_btn.bind("<Enter>", self.enter_button)
        setattr(FileExtensions, "instance_cbtn", self.checkbuttons)

    def enter_button(self, event) -> None:
        btn = event.widget
        btn.bind("<ButtonPress-1>", self.press_button)

    def press_button(self, event) -> None:
        btn = event.widget
        btn.bind("<ButtonRelease-1>", self.toggle_ext)

    @decorators.check_option_disabled
    def toggle_ext(self, event) -> None:
        btn = event.widget
        key = self.checkbuttons[btn][0]
        value = self.checkbuttons[btn][1]
        if value.get():
            self.data["file_extensions"][key] = False
        elif not value.get():
            self.data["file_extensions"][key] = True
        io_toml.dump_toml_data(FILE_PATHS.config, self.data)

        self.update_registry()

    def update_registry(self) -> None:
        io_winreg.write_all_valuex()

    def update_state(self) -> None:
        for key in self.checkbuttons.keys():
            if io_toml.load_toml_value(FILE_PATHS.config, "gui.context_menu"):
                key.state(["!disabled"])
            else:
                key.state(["disabled"])


class SubsearchOption(ttk.Labelframe):
    def __init__(self, parent) -> None:
        ttk.Labelframe.__init__(self, parent)
        self.configure(text="Subsearch Options", padding=10)
        self.data = io_toml.load_toml_data(FILE_PATHS.config)
        self.subsearch_options: dict[str, Any] = {
            "gui.context_menu": "Context menu",
            "gui.context_menu_icon": "Context menu icon",
            "gui.system_tray": "System tray icon",
            "gui.summary_notification": "Notification when done",
            "gui.show_terminal": "Terminal while searching",
            "misc.single_instance": "Single instance",
        }
        for name, description in self.subsearch_options.items():
            self.subsearch_options[name] = [
                io_toml.load_toml_value(FILE_PATHS.config, name),
                description,
            ]

        self.checkbuttons: dict[ttk.Checkbutton, tuple[str, BooleanVar]] = {}
        frame = None

        for enum, (key, value) in enumerate(self.subsearch_options.items()):
            if enum % 2 == 0:
                frame = ttk.Frame(self)
                frame.pack(side=tk.LEFT, anchor="n")

            boolean = tk.BooleanVar()
            boolean.set(value[0])
            check_btn = ttk.Checkbutton(frame, text=value[1], onvalue=True, offvalue=False, variable=boolean)
            context_menu = io_toml.load_toml_value(FILE_PATHS.config, "gui.context_menu")
            system_tray = io_toml.load_toml_value(FILE_PATHS.config, "gui.system_tray")
            if key == "gui.context_menu_icon" and not context_menu:
                check_btn.state(["disabled"])
            if key == "gui.show_terminal" and DEVICE_INFO.mode == "executable":
                check_btn.state(["disabled"])
            if key == "gui.summary_notification" and not system_tray:
                check_btn.state(["disabled"])
            check_btn.pack(padx=4, pady=4, ipadx=40)

            self.checkbuttons[check_btn] = key, boolean
            check_btn.bind("<Enter>", self.enter_button)

    def enter_button(self, event) -> None:
        btn = event.widget
        btn.bind("<ButtonPress-1>", self.press_button)

    def press_button(self, event) -> None:
        btn = event.widget
        btn.bind("<ButtonRelease-1>", self.toggle_btn)

    @decorators.check_option_disabled
    def toggle_btn(self, event) -> None:
        btn = event.widget
        key = self.checkbuttons[btn][0]
        value = self.checkbuttons[btn][1]
        if value.get():
            io_toml.update_toml_key(FILE_PATHS.config, key, False)
        elif not value.get():
            io_toml.update_toml_key(FILE_PATHS.config, key, True)
        self.update_registry(btn)
        keys = [("gui.context_menu", "gui.context_menu_icon"), ("gui.system_tray", "gui.summary_notification")]
        for key_pair in keys:
            self.disable_check_btn_children(btn, value, key_pair)

    def disable_check_btn_children(self, btn: Any, value: BooleanVar, key_pair: tuple[str, str]) -> None:
        parent_key, child_key = key_pair[0], key_pair[1]
        if btn["text"] != self.subsearch_options[parent_key][1]:
            return None
        _handle_file_extensions_check_btn(self, parent_key)
        _handle_other_check_btn(self, value, child_key)

    def update_registry(self, btn) -> None:
        if btn["text"] != self.subsearch_options["gui.context_menu"][1]:
            return None
        menu = io_toml.load_toml_value(FILE_PATHS.config, "gui.context_menu")
        if menu:
            io_winreg.add_context_menu()
        else:
            io_winreg.del_context_menu()


class DownloadManagerOptions(ttk.Labelframe):
    def __init__(self, parent) -> None:
        ttk.Labelframe.__init__(self, parent)
        self.configure(text="Download manager options", padding=10)
        self.data = io_toml.load_toml_data(FILE_PATHS.config)
        self.download_manager_options: dict[str, Any] = {
            "download_manager.open_on_no_matches": "Open on no matches found",
            "download_manager.always_open": "Always open",
            "download_manager.automatic_downloads": "Automatic downloads",
        }
        for name, description in self.download_manager_options.items():
            self.download_manager_options[name] = [
                io_toml.load_toml_value(FILE_PATHS.config, name),
                description,
            ]

        self.checkbuttons: dict[ttk.Checkbutton, tuple[str, BooleanVar]] = {}
        frame = None

        for enum, (key, value) in enumerate(self.download_manager_options.items()):
            if enum % 1 == 0:
                frame = ttk.Frame(self)
                frame.pack(side=tk.LEFT, anchor="n")

            boolean = tk.BooleanVar()
            boolean.set(value[0])
            check_btn = ttk.Checkbutton(frame, text=value[1], onvalue=True, offvalue=False, variable=boolean)
            check_btn.pack(padx=4, pady=4, ipadx=40)

            self.checkbuttons[check_btn] = key, boolean
            check_btn.bind("<Enter>", self.enter_button)

    def enter_button(self, event) -> None:
        btn = event.widget
        btn.bind("<ButtonPress-1>", self.press_button)

    def press_button(self, event) -> None:
        btn = event.widget
        btn.bind("<ButtonRelease-1>", self.toggle_btn)

    @decorators.check_option_disabled
    def toggle_btn(self, event) -> None:
        btn = event.widget
        key = self.checkbuttons[btn][0]
        value = self.checkbuttons[btn][1]
        if value.get():
            io_toml.update_toml_key(FILE_PATHS.config, key, False)
        elif not value.get():
            io_toml.update_toml_key(FILE_PATHS.config, key, True)

class AdvancedUser(ttk.Labelframe):
    def __init__(self, parent) -> None:
        ttk.Labelframe.__init__(self, parent)
        self.configure(text="Advanced user ( Requests / API )", padding=10)
        self.data = io_toml.load_toml_data(FILE_PATHS.config)
        self.tip_present = False
        self.tip = None
        adv_user = self.data["advanced_user"]
        adv_default_values = [
            adv_user["api_call_limit"],
            adv_user["request_connect_timeout"],
            adv_user["request_read_timeout"],
        ]

        self.adv_user_options: dict = {
            "advanced_user.api_call_limit": "API call limit",
            "advanced_user.request_connect_timeout": "Request connection timeout",
            "advanced_user.request_read_timeout": "Request read timeout",
        }
        for name, description in self.adv_user_options.items():
            self.adv_user_options[name] = [io_toml.load_toml_value(FILE_PATHS.config, name), description]
        frame = None
        frame_c = tk.Frame(self)
        frame_r = tk.Frame(self)

        self.entry_fields: dict[ttk.Entry, tuple[str, tk.IntVar]] = {}
        for enum, (key, value) in enumerate(self.adv_user_options.items()):
            bool_value = value[0]
            description = value[1]

            int_value = tk.IntVar()
            int_value.set(bool_value)
            field_name = ttk.Label(frame_c, text=description, justify="left")
            field = ttk.Entry(frame_c, width=3, justify="right")
            field.insert(tk.END, adv_default_values[enum])
            field_name.pack(side="left", anchor="center", fill="x")
            field.pack(side="left", anchor="center", padx=4)
            field.bind("<Enter>", self.enter_entry_field)
            self.entry_fields[field] = key, int_value
            # field.bind("<Enter>", self.enter_button)
        frame_c.pack(side=tk.LEFT, expand=True, fill="x")
        frame_r.pack(side=tk.LEFT, expand=False, after=frame_c)

        apply_input = ttk.Button(frame_r, text="Apply", width=10)
        apply_input.bind("<Enter>", self.enter_btn_apply_input)
        apply_input.bind("<Leave>", self.leave_btn_apply_input)
        apply_input.pack(anchor="center")

    def enter_btn_apply_input(self, event) -> None:
        btn = event.widget
        field_ok = True
        tip_text = (
            "If you don't know what these values do, don't touch them.",
            "Sending too many requests can get your IP banned.",
        )
        self.tip = common_utils.ToolTip(btn, btn, *tip_text)
        self.tip.show()
        self.tip_present = True
        for field, values in self.entry_fields.items():
            field: ttk.Entry
            user_input = self.verify_field(field)
            if field.instate(["invalid"]):
                self.on_invalid_state(user_input)
                btn.unbind("<ButtonPress-1>")
                field_ok = False
        if field_ok:
            btn.bind("<ButtonPress-1>", self.update_config)

    def leave_btn_apply_input(self, event) -> None:
        btn = event.widget
        self.tip.hide()
        self.tip_present = False
        btn.bind("<ButtonPress-1>", self.enter_btn_apply_input)


    def update_config(self, event) -> None:
        for field, values in self.entry_fields.items():
            value = field.get()
            key = values[0]
            io_toml.update_toml_key(FILE_PATHS.config, key, int(value))

    def verify_field(self, field: ttk.Entry) -> str:
        user_input = field.get()
        if not string_parser.valid_api_request_input(user_input):
            field.state(["invalid"])
        else:
            field.state(["!invalid"])
        return user_input

    def on_invalid_state(self, path): ...

    def enter_entry_field(self, event): ...


class CheckForUpdates(ttk.Labelframe):
    def __init__(self, parent) -> None:
        ttk.Labelframe.__init__(self, parent)
        self.configure(text="Update Subsearch", padding=10, style="TLabelframePlain")
        frame_left = tk.Frame(self, bg=cfg.color.default_bg)
        frame_right = tk.Frame(self, bg=cfg.color.default_bg)

        self.var_current = tk.StringVar(self, f"Current version:\t\t{VERSION}")
        self.var_latest = tk.StringVar(self, "Latest version: \t\t")
        self.var_misc = tk.StringVar()

        frame_current = tk.Label(frame_left, textvariable=self.var_current, justify="left")
        frame_latest = tk.Label(frame_left, textvariable=self.var_latest, justify="left")
        self.frame_misc = tk.Label(frame_left, textvariable=self.var_misc, justify="left")
        frame_current.pack(anchor="w")
        frame_latest.pack(anchor="w")
        self.frame_misc.pack(pady=16, anchor="w")

        self.btn_search = ttk.Button(
            frame_right,
            text="Search",
            width=20,
        )
        self.btn_visit_release_page = ttk.Button(
            frame_right,
            text="Open in webbrowser",
            width=20,
        )
        self.btn_update_install = ttk.Button(
            frame_right,
            text="Download & Update",
            width=20,
        )
        self.btn_search.pack(padx=2, pady=2)
        self.btn_visit_release_page.pack(padx=2, pady=2)
        self.btn_visit_release_page.state(["disabled"])
        self.btn_update_install.pack(padx=2, pady=2)
        self.btn_update_install.state(["disabled"])
        self.btn_search.bind("<Enter>", self.enter_button)

        self.pack(anchor="center", fill="x")
        width = self.winfo_reqwidth()
        half_width = round(width // 2)
        frame_left.pack(side=tk.LEFT, anchor="n", expand=True, fill="x", ipadx=half_width)
        frame_right.pack(side=tk.LEFT, anchor="n", expand=True, fill="x", ipadx=half_width)

    def enter_button(self, event) -> None:
        btn = event.widget
        if btn == self.btn_search:
            btn.bind("<ButtonRelease-1>", self.search_button)

    def search_button(self, event) -> None:
        new_repo_avail, repo_is_prerelease = update.is_new_version_avail()
        latest_version = update.get_latest_version()
        if new_repo_avail:
            self.var_latest.set(f"Latest version: \t\t{latest_version}")
            self.btn_search.state(["disabled"])
            self.btn_visit_release_page.state(["!disabled"])
            self.btn_visit_release_page.bind("<ButtonRelease-1>", self.visit_repository_button)
            if DEVICE_INFO.mode == "executable":
                self.btn_update_install.state(["!disabled"])
                self.btn_update_install.bind("<ButtonRelease-1>", self.download_and_update)
            else:
                self.btn_update_install["text"] = "Only availible with MSI"
            if repo_is_prerelease:
                self.var_misc.set(f"Pre-release")
                self.frame_misc.configure(fg=cfg.color.orange)
            else:
                self.var_misc.set(f"Latest")
                self.frame_misc.configure(fg=cfg.color.green)

        if not new_repo_avail:
            self.var_latest.set(f"Latest version: \t\t{latest_version}")
            self.var_misc.set(f"Latest version already installed")

    def visit_repository_button(self, event) -> None:
        webbrowser.open(f"https://github.com/vagabondHustler/subsearch/releases")

    def download_and_update(self, event) -> None:
        update.download_and_update()
        self.btn_visit_release_page.state(["disabled"])



import shutil
import struct
import time
from typing import Iterable, Optional
import zipfile
from io import BufferedReader
from pathlib import Path

import requests

from subsearch.globals import log
from subsearch.globals.constants import VIDEO_FILE
from subsearch.globals.dataclasses import Subtitle
from subsearch.providers.common_utils import get_cloudscraper
from subsearch.utils import string_parser


def create_path_from_string(string: str, path_resolution: str) -> Path:
    if path_resolution == "relative":
        if string == ".":
            path = VIDEO_FILE.file_directory
        elif string.startswith(".\\") and len(string) > 2:
            path = VIDEO_FILE.file_directory.joinpath(string[2:])
        elif string == "..":
            path = VIDEO_FILE.file_directory.parent
        elif string.startswith("..\\") and len(string) > 3:
            path = VIDEO_FILE.file_directory.parent.joinpath(string[3:])
    elif path_resolution == "absolute":
        path = Path(string)

    path.mkdir(parents=True, exist_ok=True)
    return path


def download_subtitle(subtitle: Subtitle, index_position: int, index_size: int) -> None:
    log.stdout(f"{subtitle.provider_name}: {index_position}/{index_size}: {subtitle.subtitle_name}")
    scraper = get_cloudscraper()
    r = scraper.get(subtitle.download_url, stream=True)
    file_name = f"{subtitle.provider_name}_{subtitle.subtitle_name}_{index_position}.zip"
    download_path = VIDEO_FILE.tmp_dir / file_name
    with open(download_path, "wb") as fd:
        for chunk in r.iter_content(chunk_size=1024):
            fd.write(chunk)


def extract_files_in_dir(src: Path, dst: Path, extension: str = ".zip") -> None:
    for file in src.glob(f"*{extension}"):
        filename = src / file
        log.file_system_action(action_type="extract", src=file, dst=dst)
        zip_ref = zipfile.ZipFile(filename)
        zip_ref.extractall(dst)
        zip_ref.close()


def autoload_rename(release_name: str, extension: str = ".srt") -> Path:
    best_match = (0, Path("."))
    for file in VIDEO_FILE.subs_dir.glob(f"*{extension}"):
        value = string_parser.calculate_match(file.name, release_name)
        if value >= best_match[0]:
            best_match = value, file

    old_file_path = best_match[1]
    new_file_path = old_file_path.with_name(f"{release_name}{extension}")
    log.file_system_action(action_type="rename", src=old_file_path, dst=new_file_path)
    old_file_path.rename(new_file_path)
    return new_file_path


def move_all(src: Path, dst: Path, extension: str = ".srt") -> None:
    for file in src.glob(f"*{extension}"):
        move_and_replace(file.absolute(), dst)


def move_and_replace(source_file: Path, destination_directory: Path) -> None:
    source_file.replace(destination_directory / source_file.name)
    log.file_system_action(action_type="move", src=source_file, dst=destination_directory)


def del_file_type(cwd: Path, extension: str) -> None:
    for file in Path(cwd).glob(f"*{extension}"):
        log.file_system_action(action_type="remove", src=file)
        file_path = Path(cwd) / file
        file_path.unlink()


def del_directory(directory: Path) -> None:
    log.file_system_action(action_type="remove", src=directory)
    shutil.rmtree(directory)


def directory_is_empty(directory: Path) -> bool:
    if not any(directory.iterdir()):
        return True
    return False


def del_directory_content(directory: Path) -> None:
    for item in directory.iterdir():
        if not item.is_file():
            return None
        log.file_system_action(action_type="remove", src=item)
        if item.is_file():
            item.unlink()
        if item.is_dir():
            shutil.rmtree(item)


def create_directory(path: Path) -> None:
    if path.exists():
        return None
    log.stdout(f"Creating {path}", level="debug")
    path.mkdir(parents=True, exist_ok=True)


def get_file_hash(file_path: Path) -> str:
    log.stdout("Calculating hash of video file", level="debug")
    if not file_path:
        return ""

    hash_algorithm = MPCHashAlgorithm(file_path)
    return hash_algorithm.get_hash()


def count_files_in_directory(path: Path, extensions: Optional[Iterable[str]] = None) -> int:
    extensions = {ext.lower() for ext in extensions} if extensions else None
    file_count = sum(
        1 for file in path.iterdir() if file.is_file() and (extensions is None or file.suffix.lower() in extensions)
    )
    return file_count


class MPCHashAlgorithm:
    def __init__(self, file_path: Path, **kwargs) -> None:
        file_size = file_path.stat().st_size
        self.file_path = file_path
        self.chunk_size = kwargs.get("chunk_size", 65536)
        self.file_size, self.hash_chunk = file_size, file_size
        self.hash = self.mpc_hash_algorithm()

    def mpc_hash_algorithm(self) -> str:
        if not self.valid_file_size():
            return ""

        with open(self.file_path, "rb") as file:
            self.byte_size = struct.calcsize("<q")
            self.update_hash_chunk(file)
            file.seek(max(0, self.file_size - self.chunk_size), 0)
            self.update_hash_chunk(file)
        return "%016x" % self.hash_chunk

    def valid_file_size(self) -> bool:
        if self.file_size < self.chunk_size * 2:
            log.stdout(f"Invalid file size, {self.file_size} bytes", level="warning")
            return False
        return True

    def update_hash_chunk(self, file: BufferedReader) -> None:
        for _ in range(self.chunk_size // self.byte_size):
            buffer = file.read(self.byte_size)
            (chunk_value,) = struct.unpack("<q", buffer)
            self.hash_chunk += chunk_value
            self.hash_chunk = self.hash_chunk & 0xFFFFFFFFFFFFFFFF

    def get_hash(self) -> str:
        return self.hash


def download_response(msi_package_path: Path, response: requests.Response) -> None:
    start_time = time.time()
    with open(msi_package_path, "wb") as msi_file:
        total_size = int(response.headers.get("content-length", 0))
        downloaded_size = 0
        log.stdout(f"Download started for {msi_package_path.name}")
        log.stdout(f"Downloading 0%")
        for chunk in response.iter_content(chunk_size=128):
            msi_file.write(chunk)
            downloaded_size += len(chunk)
            progress_percentage = (downloaded_size / total_size) * 100
            elapsed_time = time.time() - start_time
            if elapsed_time >= 0.5:
                log.stdout(f"Downloading {progress_percentage:.2f}%")
                start_time = time.time()
        log.stdout(f"Download complete.")



from PIL import Image, ImageTk

from subsearch.globals.constants import APP_PATHS
from subsearch.gui import sprites


def get_sprite(sprite_name) -> Image.Image:
    spritesheet_path = APP_PATHS.gui_assets / "spritesheet.png"
    spritesheet_image = Image.open(spritesheet_path)
    sprite_x = sprites[sprite_name][0]
    sprite_y = sprites[sprite_name][1]
    sprite_width = sprites[sprite_name][2]
    sprite_height = sprites[sprite_name][3]
    sprite = spritesheet_image.crop((sprite_x, sprite_y, sprite_x + sprite_width, sprite_y + sprite_height))
    return sprite


def asset_menu_btn(cls, img, type, x=27, y=27) -> None:
    path = get_sprite(f"{img}_{type}")
    png = ImageTk.PhotoImage(path)
    update_asset(cls, png, x, y)


def update_asset(cls, img, x, y) -> None:
    cls.delete("all")
    cls.create_image(x, y, image=img)
    cls.photoimage = img


def set_ttk_theme(root) -> None:
    initializer_tcl = APP_PATHS.gui_styles / "theme_setter.tcl"
    root.tk.call("source", str(initializer_tcl))
    root.tk.call("set_theme")



import tkinter as tk
from tkinter import ttk

from subsearch.globals import log
from subsearch.globals.constants import FILE_PATHS, VIDEO_FILE
from subsearch.globals.dataclasses import Subtitle
from subsearch.gui import common_utils
from subsearch.gui.resources import config as cfg
from subsearch.providers import subsource
from subsearch.utils import io_file_system, io_toml, string_parser


class DownloadManager(ttk.LabelFrame):
    downloaded_subtitle: list[Subtitle] = []

    def __init__(self, parent, **kwargs) -> None:
        ttk.Labelframe.__init__(self, parent)
        self.configure(text="Available subtitles", padding=10)
        subtitles: list[Subtitle] = kwargs.get("subtitles", [])
        if subtitles:
            subtitles.sort(key=lambda x: x.precentage_result, reverse=True)
        frame_left = tk.Frame(self)
        frame_sep = tk.Frame(self, width=5)
        frame_right = tk.Frame(self)
        frame_left.pack(side=tk.LEFT, expand=True, fill="both")
        frame_sep.pack(side=tk.LEFT, fill="y")
        frame_right.pack(side=tk.LEFT, fill="y")
        self.failed_subtitle_downloads: list[Subtitle] = []
        self.download_number = 1
        self.download_index_size = len(subtitles)
        self.subtitles = subtitles
        self.scrollbar = ttk.Scrollbar(frame_right, orient="vertical", style="Vertical.TScrollbar")
        self.sub_listbox = tk.Listbox(
            frame_left,
            height=36,
            bg=cfg.color.default_bg,
            fg=cfg.color.light_grey,
            font=cfg.font.cas8b,
            bd=0,
            border=0,
            borderwidth=0,
            highlightthickness=0,
            activestyle="none",
            yscrollcommand=self.scrollbar.set,
        )
        self.sub_listbox.pack(expand=True, fill="both")
        if self.subtitles is not None:
            self.fill_listbox()
        self.scrollbar.pack(expand=True, fill="y")
        self.scrollbar.config(command=self.sub_listbox.yview)
        if not self.subtitles:
            self.nothing_label = tk.Label(
                self,
                text="It's very empty...\n\n:(",
                font=cfg.font.cas20b,
                fg=cfg.color.default_fg,
                bg=cfg.color.default_bg,
                anchor="center",
                justify="center",
            )
            self.nothing_label.place(x=cfg.size.height // 2, y=cfg.size.width // 4, anchor="center")

    def fill_listbox(self) -> None:
        accept_threshold = io_toml.load_toml_value(FILE_PATHS.config, "subtitle_filters.accept_threshold")
        automatic_downloads = io_toml.load_toml_value(FILE_PATHS.config, "download_manager.automatic_downloads")
        self.listbox_index: dict[int, Subtitle] = {}
        for enum, subtitle in enumerate(self.subtitles):
            self.sub_listbox.insert(tk.END, f"{subtitle.precentage_result}% {subtitle.subtitle_name}\n")
            if subtitle.precentage_result == accept_threshold and not automatic_downloads:
                self.downloaded_subtitle.append(subtitle)
                self.download_number += 1
                self.update_text(enum, "✓", subtitle, cfg.color.green)
            self.sub_listbox.bind("<ButtonPress-1>", self.mouse_b1_press)
            self.listbox_index[enum] = subtitle

    def mouse_b1_press(self, event) -> None:
        self.sub_listbox.unbind("<ButtonPress-1>")
        self.sub_listbox.bind("<<ListboxSelect>>", self.select_subtitle)

    def select_subtitle(self, event) -> None:
        self.sub_listbox.unbind("<<ListboxSelect>>")
        _selection = self.sub_listbox.curselection()
        selection = _selection[0]
        subtitle = self.listbox_index[selection]
        if subtitle in (self.downloaded_subtitle or self.failed_subtitle_downloads):
            self.sub_listbox.bind("<ButtonPress-1>", self.mouse_b1_press)
            return
        if subtitle.provider_name == "subsource":
            subsource_api = subsource.GetDownloadUrl()
            download_url = subsource_api.get_url(subtitle)
            if not download_url:
                self.update_text(selection, "⨯", subtitle, cfg.color.red)
            else:
                subtitle.download_url = download_url
                subtitle.request_data = {}
        self.update_text(selection, "⊙", subtitle, cfg.color.orange)
        self.sub_listbox.bind("<ButtonRelease-1>", lambda event: self.download(event, subtitle, selection))

    def download(self, event, subtitle: Subtitle, selection: int) -> None:
        self.sub_listbox.unbind("<ButtonRelease-1>")
        try:
            if string_parser.valid_filename(subtitle.subtitle_name):
                subtitle.subtitle_name = string_parser.fix_filename(subtitle.subtitle_name)
            io_file_system.download_subtitle(subtitle, self.download_number, self.download_index_size)
            io_file_system.extract_files_in_dir(VIDEO_FILE.tmp_dir, VIDEO_FILE.subs_dir)
            zip_archive = (
                VIDEO_FILE.tmp_dir / f"{subtitle.provider_name}_{subtitle.subtitle_name}_{self.download_number}.zip"
            )
            zip_archive.unlink()
            self.update_text(selection, "✓", subtitle, cfg.color.green)
            self.download_number += 1
            self.download_index_size += 1
            self.downloaded_subtitle.append(subtitle)
        except Exception as e:
            log.stdout(str(e), level="error")
            self.update_text(selection, "⨯", subtitle, cfg.color.red)
            self.failed_subtitle_downloads.append(subtitle)
        finally:
            self.sub_listbox.bind("<ButtonPress-1>", self.mouse_b1_press)

    def update_text(self, selection: int, symbol: str, subtitle: Subtitle, color: str) -> None:
        self.sub_listbox.delete(int(selection))
        self.sub_listbox.insert(int(selection), f"{symbol} {subtitle.subtitle_name}")
        self.sub_listbox.itemconfig(int(selection), {"fg": color})



from dataclasses import Field, dataclass
from pathlib import Path
from typing import Any, ClassVar, Protocol


@dataclass(order=True)
class LanguageData:
    name: str
    alpha_1: str
    alpha_2b: str
    incompatibility: list[str]


@dataclass(order=True, slots=True)
class ProviderAlphaCodeData:
    provider: str
    alpha_code: str


@dataclass(order=True, slots=True)
class AppPaths:
    home: Path
    data: Path
    gui: Path
    gui_assets: Path
    gui_styles: Path
    providers: Path
    utils: Path
    tmp_dir: Path
    appdata_subsearch: Path


@dataclass(order=True, slots=True)
class FilePaths:
    log: Path
    config: Path
    language_data: Path


@dataclass(order=True, slots=True)
class VideoFile:
    file_exist: bool
    filename: str
    file_hash: str
    file_extension: str
    file_path: Path
    file_directory: Path
    subs_dir: Path
    tmp_dir: Path


@dataclass(order=True, slots=True)
class AppConfig:
    current_language: str
    accept_threshold: int
    hearing_impaired: bool
    non_hearing_impaired: bool
    only_foreign_parts: bool
    context_menu: bool
    context_menu_icon: bool
    system_tray: bool
    summary_notification: bool
    show_terminal: bool
    subtitle_post_processing: dict[str, Any]
    file_extensions: dict[str, bool]
    providers: dict[str, bool]
    open_on_no_matches: bool
    always_open: bool
    automatic_downloads: bool
    single_instance: bool
    api_call_limit: int
    request_connect_timeout: int
    request_read_timeout: int


@dataclass(order=True, slots=True)
class ReleaseData:
    title: str
    year: int
    season: str
    season_ordinal: str
    episode: str
    episode_ordinal: str
    tvseries: bool
    release: str
    group: str
    imdb_id: str


@dataclass(order=True, slots=True)
class ProviderUrls:
    opensubtitles: str
    opensubtitles_hash: str
    yifysubtitles: str
    subsource: str


@dataclass(order=True, slots=True)
class Subtitle:
    precentage_result: int
    provider_name: str
    subtitle_name: str
    download_url: str
    request_data: dict[str, Any]


@dataclass(order=True, frozen=True)
class SystemInfo:
    platform: str
    mode: str
    python: str
    subsearch: str


@dataclass(order=True, slots=True, frozen=True)
class WindowsRegistryPaths:
    classes: str
    asterisk: str
    shell: str
    subsearch: str
    subsearch_command: str
    long_paths: str


class GenericDataClass(Protocol):
    __dataclass_fields__: ClassVar[dict[str, Field[Any]]]


@dataclass(order=True, slots=True)
class ProviderApiLimit:
    opensubtitles: int
    yifysubtitles: int
    subsource: int



import re
import subprocess
import sys
from pathlib import Path

import cloudscraper
import requests
from packaging.version import Version

from subsearch.globals import log
from subsearch.globals.constants import APP_PATHS, VERSION
from subsearch.utils import io_file_system


def find_semantic_version(version: str) -> str:
    # semantic expression https://regex101.com/r/M4qItH/
    pattern = r'(?<=__version__ = ")(\d+\.\d+\.\d+[a-zA-Z]*\d*).*?(?=")'
    version_semantic = "".join(re.findall(pattern, version)[0])
    return version_semantic


def scrape_github() -> str:
    scraper = cloudscraper.create_scraper(browser={"browser": "chrome", "platform": "android", "desktop": False})
    url = "https://raw.githubusercontent.com/vagabondHustler/subsearch/main/src/subsearch/data/version.py"
    source = scraper.get(url)
    scontent = source.content
    file_content = str(scontent)
    return file_content


def get_latest_version() -> str:
    version_github = scrape_github()
    return find_semantic_version(version_github)


def is_new_version_avail() -> tuple[bool, bool]:
    new_repo_avail, repo_is_prerelease = False, False
    released_version = get_latest_version()

    if Version(VERSION) < Version(released_version):
        if Version(released_version).is_prerelease:
            repo_is_prerelease = True
        new_repo_avail = True

    return new_repo_avail, repo_is_prerelease


def get_latest_msi_url(latest_version: str = "") -> str:
    return f"https://github.com/vagabondHustler/subsearch/releases/download/{latest_version}/Subsearch-{latest_version}-win64.msi"


def run_installer(msi_package_path: Path) -> None:
    command = f"msiexec.exe /i {msi_package_path}"
    subprocess.Popen(command, shell=True, creationflags=subprocess.DETACHED_PROCESS)


def download_and_update() -> None:
    log.brackets("Updating Application")
    latest_version = get_latest_version()
    latest_msi = get_latest_msi_url(latest_version)
    if Version(VERSION) > Version(latest_version):
        log.stdout(f"No new version available")
        return None

    log.stdout(f"New version available")
    if not APP_PATHS.tmp_dir.exists():
        APP_PATHS.tmp_dir.mkdir(parents=True, exist_ok=True)
    msi_package_path = APP_PATHS.tmp_dir / f"Subsearch-{latest_version}-win64.msi"
    response = requests.get(latest_msi, stream=True)
    if response.status_code == 200:
        io_file_system.download_response(msi_package_path, response)
        msg = f"MSI file downloaded to: {msi_package_path}"
        log.stdout(f"MSI file downloaded to: {msi_package_path}")
    else:
        msg = f"Failed to download MSI file. HTTP Status Code: {response.status_code}"
        log.stdout(msg, level="error")
        raise Exception(response.status_code)
    run_installer(msi_package_path)
    sys.exit()



import functools
from pathlib import Path
from typing import Any, Union

import toml

from subsearch.globals import log
from subsearch.globals.constants import DEFAULT_CONFIG, FILE_PATHS
from subsearch.globals.dataclasses import AppConfig


def load_toml_data(toml_file_path: Path) -> dict[str, Any]:
    with open(toml_file_path, "r") as f:
        return toml.load(f)


def load_toml_value(toml_file_path: Path, key: str) -> Any:
    toml_data = load_toml_data(toml_file_path)
    keys = key.split(".")
    value = functools.reduce(dict.get, keys, toml_data)  # type: ignore
    return value


def dump_toml_data(toml_file_path: Path, toml_data: dict) -> None:
    with open(toml_file_path, "w") as f:
        toml.dump(toml_data, f)


def update_toml_key(toml_file_path: Path, key: str, value: Any | None) -> None:
    toml_data = load_toml_data(toml_file_path)
    keys = key.split(".")
    functools.reduce(dict.get, keys[:-1], toml_data)[keys[-1]] = value  # type: ignore
    dump_toml_data(toml_file_path, toml_data)


def del_toml_key(toml_file_path: Path, key: str) -> None:
    toml_data = load_toml_data(toml_file_path)
    keys = key.split(".")
    functools.reduce(dict.get, keys[:-1], toml_data).pop(keys[-1], None)  # type: ignore
    dump_toml_data(toml_file_path, toml_data)


def repair_toml_config(toml_file_path: Path, valid_config_keys: list[str], config_keys: list[str]) -> None:
    log.stdout(f"Reparing config", level="debug")
    obsolete_keys = [key for key in config_keys if key not in valid_config_keys]
    for key in obsolete_keys:
        log.stdout(f"Removing obsolete key {key}", level="debug")
        del_toml_key(toml_file_path, key)

    missing_keys = [key for key in valid_config_keys if key not in config_keys]
    for key in missing_keys:
        log.stdout(f"Adding missing key {key}", level="debug")
        keys = key.split(".")
        value = functools.reduce(dict.get, keys, DEFAULT_CONFIG)  # type: ignore
        update_toml_key(FILE_PATHS.config, key, value)


def get_keys_recursively(dictionary: dict, prefix="", keys=None) -> list[str]:
    if keys is None:
        keys = []

    if isinstance(dictionary, dict):
        for key in dictionary:
            full_key = f"{prefix}.{key}" if prefix else key
            keys.append(full_key)
            get_keys_recursively(dictionary[key], full_key, keys)

    return keys


def valid_config(valid_config_keys: list[str], config_keys: list[str]) -> bool:
    if not FILE_PATHS.config.exists():
        return False
    valid_config_keys.sort()
    config_keys.sort()
    return config_keys == valid_config_keys


def resolve_on_integrity_failure() -> None:
    if not FILE_PATHS.config.exists():
        log.stdout(f"No config.toml exsist, creating default at {FILE_PATHS.config}", level="debug")
        dump_toml_data(FILE_PATHS.config, DEFAULT_CONFIG)
        return None
    valid_config_keys = get_keys_recursively(DEFAULT_CONFIG)
    config_keys = get_keys_recursively(load_toml_data(FILE_PATHS.config))
    valid = valid_config(valid_config_keys, config_keys)
    if valid:
        log.stdout(f"Config is valid", level="debug")
        return None
    try:
        log.stdout(f"Config not valid", level="debug")
        repair_toml_config(FILE_PATHS.config, valid_config_keys, config_keys)
    except Exception:
        log.stdout(f"Repair faild, creating default at {FILE_PATHS.config}", level="debug")
        FILE_PATHS.config.unlink()
        dump_toml_data(FILE_PATHS.config, DEFAULT_CONFIG)
    log.stdout(f"Repair succeeded", level="debug")


def get_app_config(toml_file_path: Path) -> AppConfig:
    data = load_toml_data(toml_file_path)
    user_data = AppConfig(
        **data["subtitle_filters"],
        **data["gui"],
        subtitle_post_processing=data["subtitle_post_processing"],
        file_extensions=data["file_extensions"],
        providers=data["providers"],
        **data["download_manager"],
        **data["misc"],
        **data["advanced_user"],
    )
    return user_data



import re
from typing import Any

from num2words import num2words

from subsearch.globals import log
from subsearch.globals.constants import VIDEO_FILE
from subsearch.globals.dataclasses import (
    AppConfig,
    LanguageData,
    ProviderUrls,
    ReleaseData,
)
from subsearch.utils import imdb_lookup


def remove_padded_zero(x: str) -> str:
    lenght = len(x)
    if x.startswith("0") and lenght > 1:
        return str(x)[1:]
    return x


def find_year(string: str) -> int:
    re_year = re.findall(r"^.*\.([1-2][0-9]{3})\.", string)
    if re_year:
        year = re_year[0]
        return int(year)
    return 0000


def find_title_by_year(string: str) -> str:
    re_title = re.findall(r"^(.*)\.[1-2][0-9]{3}\.", string)
    if re_title:
        title: str = re_title[0]
        title = title.replace(".", " ")
        return title
    return ""


def find_title_by_show(string: str) -> str:
    re_title = re.findall(r"^(.*)\.[s]\d*[e]\d*\.", string)
    if re_title:
        title: str = re_title[0]
        title = title.replace(".", " ")
        return title
    return ""


def find_season_episode(string: str) -> str:
    re_se = re.findall(r"\.([s]\d*[e]\d*)\.", string)
    if re_se:
        se: str = re_se[0]
        return se
    return ""


def convert_to_ordinal_string(string: str) -> tuple[str, str, str, str, bool]:
    if string == "":
        season, season_ordinal, episode, episode_ordinal = "", "", "", ""
        show_bool = False
    else:
        season, episode = string.replace("s", "").replace("e", " ").split(" ")
        season_ordinal = num2words(int(season), lang="en", to="ordinal")
        episode_ordinal = num2words(int(episode), lang="en", to="ordinal")
        show_bool = True
    return season, season_ordinal, episode, episode_ordinal, show_bool


def find_group(string: str) -> str:
    group = string.rsplit("-", 1)[-1]
    return group


def find_title(filename: str, year: int, series: bool) -> str:
    if year != 0000:
        title = find_title_by_year(filename)
    elif series and year == 0000:
        title = find_title_by_show(filename)
    else:
        title = filename.rsplit("-", 1)[0]
    return title


class CreateProviderUrls:
    def __init__(self, app_config: AppConfig, release_data: ReleaseData, language_data: dict[str, Any]) -> None:
        self.app_config = app_config
        self.release_data = release_data
        self.language_data = language_data
        self.current_language_data: LanguageData = LanguageData(**language_data[app_config.current_language])

    def retrieve_urls(self) -> ProviderUrls:
        urls = ProviderUrls(
            opensubtitles=self.opensubtitles,
            opensubtitles_hash=self.opensubtitles_hash,
            yifysubtitles=self.yifysubtitles,
            subsource=self.subsource,
        )
        return urls

    @classmethod
    def no_urls(cls) -> ProviderUrls:
        return ProviderUrls("", "", "", "")

    @property
    def subsource(self) -> str:
        return "https://api.subsource.net/api"

    @property
    def opensubtitles(self) -> str:
        domain = "https://www.opensubtitles.org"
        subtitle_type = self._opensubtitles_subtitle_type()
        search_parameter = self._opensubtitles_search_parameter()
        return f"{domain}/{subtitle_type}/{search_parameter}/rss_2_00".replace(" ", "%20")

    @property
    def opensubtitles_hash(self) -> str:
        domain = "https://www.opensubtitles.org"
        subtitle_type = self._opensubtitles_subtitle_type()
        return f"{domain}/{subtitle_type}/moviehash-{VIDEO_FILE.file_hash}"

    @property
    def yifysubtitles(self) -> str:
        if self.release_data.tvseries:
            return ""
        domain = "https://yifysubtitles.org"
        if self.release_data.imdb_id:
            return f"{domain}/movie-imdb/{self.release_data.imdb_id}"
        return ""

    def _opensubtitles_subtitle_type(self) -> str:
        alpha_2b = self.current_language_data.alpha_2b
        hearing_imparied, hearing_imparied_foreign_parts, foreign_parts = self.subtitle_type_logic()
        if hearing_imparied:
            return f"en/search/sublanguageid-{alpha_2b}/hearingimpaired-on"
        elif foreign_parts:
            return f"en/search/sublanguageid-{alpha_2b}/foreignpartsonly-on"
        elif hearing_imparied_foreign_parts:
            return f"en/search/sublanguageid-{alpha_2b}/hearingimpaired-on/foreignpartsonly-on"
        else:
            return f"en/search/sublanguageid-{alpha_2b}"

    def _opensubtitles_search_parameter(self) -> str:
        if self.release_data.tvseries:
            return f"searchonlytvseries-on/season-{self.release_data.season}/episode-{self.release_data.episode}/moviename-{self.release_data.title}"
        return f"searchonlymovies-on/moviename-{self.release_data.title} ({self.release_data.year})"

    def subtitle_type_logic(self) -> tuple[bool, bool, bool]:
        hi_all_parts = (
            self.app_config.hearing_impaired
            and not self.app_config.non_hearing_impaired
            and not self.app_config.only_foreign_parts
        )
        hi_foreign_parts = (
            self.app_config.hearing_impaired
            and not self.app_config.non_hearing_impaired
            and self.app_config.only_foreign_parts
        )
        non_hi_foreign_parts = (
            not self.app_config.hearing_impaired
            and self.app_config.non_hearing_impaired
            and self.app_config.only_foreign_parts
        )
        return hi_all_parts, hi_foreign_parts, non_hi_foreign_parts


def get_release_data(filename: str) -> ReleaseData:
    if not filename:
        return no_release_data()
    release = filename.lower()
    year = find_year(release)
    season_episode = find_season_episode(release)
    season, season_ordinal, episode, episode_ordinal, series = convert_to_ordinal_string(season_episode)

    title = find_title(release, year, series)
    group = find_group(release)
    imdb_id = ""

    parameters = ReleaseData(
        title,
        year,
        season,
        season_ordinal,
        episode,
        episode_ordinal,
        series,
        release,
        group,
        imdb_id,
    )
    return parameters


def no_release_data() -> ReleaseData:
    return ReleaseData("", 0, "", "", "", "", False, "", "", "")


def calculate_match(from_user: str, from_website: str) -> int:
    max_percentage = 100
    _from_user: list[str] = mk_lst(from_user)
    _from_website: list[str] = mk_lst(from_website)
    lst1, lst2 = make_equal_size(_from_user, _from_website)
    not_matching = list(set(lst1) - set(lst2))
    not_matching_value = len(not_matching)
    number_of_items = len(lst1)
    percentage_to_remove = round(not_matching_value / number_of_items * max_percentage)
    pct = round(max_percentage - percentage_to_remove)

    return pct


def mk_lst(release: str) -> list[str]:
    new: list[str] = []
    qualities = ["720p", "1080p", "1440p", "2160p"]
    temp = release.split(".")

    for item in temp:
        if item not in qualities:
            new.append(item.lower())
    return new


def make_equal_size(lst1, lst2) -> tuple:
    if len(lst1) == len(lst2):
        return lst1, lst2
    elif len(lst1) > len(lst2):
        lst_big, lst_small = lst1, lst2
    else:
        lst_big, lst_small = lst2, lst1

    num_big, num_small = sorted((len(lst1), len(lst2)), reverse=True)
    difference = num_big - num_small
    filled_list = fill_shorter_list(lst_big, lst_small, difference)
    return lst_big, filled_list


def fill_shorter_list(big_lst, small_lst, difference) -> Any:
    if big_lst > small_lst:
        for _i in range(difference):
            small_lst.append(None)
    return small_lst


def valid_filename(input_string) -> bool:
    forbidden_characters_pattern = r'[<>:"/\\|?*\x00]'
    return bool(re.search(forbidden_characters_pattern, input_string))


def fix_filename(input_string) -> str:
    forbidden_characters_pattern = r'[<>:"/\\|?*\x00]'
    return re.sub(forbidden_characters_pattern, ".", input_string)


def valid_path(input_str, path_resolution) -> bool:
    if input_str == "":
        return False
    if path_resolution == "relative":
        pattern = r"^\.{1,2}\\([a-z0-9-_]|\\[a-z0-9-_])+$|^\.{1,2}$"
    elif path_resolution == "absolute":
        pattern = r"^[a-zA-Z]{1}:\\([a-z0-9-_]|\\[a-z0-9-_])+$"
    return bool(re.match(pattern, input_str))


def valid_api_request_input(input: int) -> bool:
    pattern = r"[0-9]"
    return bool(re.match(pattern, input))



import platform
import socket
import sys
import tempfile
from itertools import product
from pathlib import Path
from typing import Any, no_type_check

from subsearch.data import __guid__, __version__
from subsearch.globals.dataclasses import (
    AppPaths,
    FilePaths,
    SystemInfo,
    VideoFile,
    WindowsRegistryPaths,
)


def get_app_paths() -> AppPaths:
    app_home = Path(__file__).resolve().parent.parent
    return AppPaths(
        home=app_home,
        data=app_home / "data",
        gui=app_home / "gui",
        gui_assets=app_home / "gui" / "resources" / "assets",
        gui_styles=app_home / "gui" / "resources" / "styles",
        providers=app_home / "providers",
        utils=app_home / "utils",
        tmp_dir=Path(tempfile.gettempdir()) / f"tmp_subsearch",
        appdata_subsearch=Path.home() / "AppData" / "Local" / "Subsearch",
    )


def get_file_paths() -> FilePaths:
    app_home = Path(__file__).resolve().parent.parent
    return FilePaths(
        log=Path.home() / "AppData" / "Local" / "Subsearch" / "log.log",
        config=Path.home() / "AppData" / "Local" / "Subsearch" / "config.toml",
        language_data=app_home / "data" / "language_data.toml",
    )


def get_default_app_config() -> dict[str, Any]:
    file_extensions = dict.fromkeys(get_supported_file_ext(), True)
    providers = dict.fromkeys(get_supported_providers(), True)
    config = {
        "subtitle_filters": {
            "current_language": "english",
            "accept_threshold": 90,
            "hearing_impaired": True,
            "non_hearing_impaired": True,
            "only_foreign_parts": False,
        },
        "gui": {
            "context_menu": True,
            "context_menu_icon": True,
            "system_tray": True,
            "summary_notification": False,
            "show_terminal": False,
        },
        "subtitle_post_processing": {
            "rename": True,
            "move_best": True,
            "move_all": False,
            "target_path": ".",
            "path_resolution": "relative",
        },
        "file_extensions": file_extensions,
        "providers": providers,
        "download_manager": {
            "open_on_no_matches": True,
            "always_open": False,
            "automatic_downloads": True,
        },
        "misc": {
            "single_instance": True,
        },
        "advanced_user": {
            "api_call_limit": 4,
            "request_connect_timeout": 4,
            "request_read_timeout": 5,
        },
    }
    return config


@no_type_check
def get_video_file_data() -> VideoFile:
    file_exist = False
    supported_exts = get_supported_file_ext()
    file_name, file_hash, file_ext = "", "", ""
    file_path = Path("")
    file_directory = Path("")
    subs_dir = Path("")
    tmp_dir = Path("")
    for i in product(supported_exts, sys.argv):
        if i[1].endswith(i[0]) and str(i[1])[i[1].rfind("\\") :].startswith("\\"):
            file_path = Path(i[1])
            file_directory = file_path.parent
            tmp_dir = file_directory / "tmp_subsearch"
            subs_dir = file_directory / "subs"
            file_name = file_path.stem
            file_ext = file_path.suffix
            file_exist = True
            file_hash = ""
            break

    video_file = VideoFile(
        file_exist=file_exist,
        filename=file_name,
        file_hash=file_hash,
        file_extension=file_ext,
        file_path=file_path,
        file_directory=file_directory,
        subs_dir=subs_dir,
        tmp_dir=tmp_dir,
    )
    return video_file


def get_system_info() -> SystemInfo:
    if getattr(sys, "frozen", False):
        mode = "executable"
        python = ""
    else:
        python = platform.python_version()
        mode = "interpreter"
    platform_ = platform.platform().lower()

    return SystemInfo(platform_, mode, python, __version__)


def get_supported_file_ext() -> list[str]:
    exts = [
        "avi",
        "mp4",
        "mkv",
        "mpg",
        "mpeg",
        "mov",
        "rm",
        "vob",
        "wmv",
        "flv",
        "3gp",
        "3g2",
        "swf",
        "mswmm",
    ]
    return exts


def get_supported_providers() -> list[str]:
    providers = ["opensubtitles_site", "opensubtitles_hash", "yifysubtitles_site", "subsource_site"]
    return providers


def get_windows_registry_paths() -> WindowsRegistryPaths:
    registry_paths = WindowsRegistryPaths(
        classes=r"Software\Classes",
        asterisk=r"Software\Classes\*",
        shell=r"Software\Classes\*\shell",
        subsearch=r"Software\Classes\*\shell\Subsearch",
        subsearch_command=r"Software\Classes\*\shell\Subsearch\command",
        long_paths=r"SYSTEM\CurrentControlSet\Control\FileSystem",
    )
    return registry_paths


def get_computer_name() -> str:
    return socket.gethostname()


def get_app_version() -> str:
    return str(__version__)


def get_guid() -> str:
    return str(__guid__)



from typing import Any

import cloudscraper
import requests

from subsearch.globals import log
from subsearch.globals.dataclasses import Subtitle
from subsearch.providers import common_utils


class SubsourceApi:
    def __init__(self, *args, **kwargs) -> None:
        self.method = {
            "search_movie": f"https://api.subsource.net/api/searchMovie",
            "get_movie": f"https://api.subsource.net/api/getMovie",
            "get_sub": f"https://api.subsource.net/api/getSub",
            "download_sub": f"https://api.subsource.net/api/downloadSub",
        }

        self.post = cloudscraper.requests.post
        self.call_limit = 5
        self.provider_name = ""

    def search_movie(self, imdb_id: str, *args, **kwargs) -> requests.Response:
        url = self.method["search_movie"]
        data = {"query": imdb_id}
        response = self.post(url=url, data=data)
        return response

    def get_movie(self, lang: str, movie_name: str, *args, **kwargs) -> requests.Response:
        url = self.method["get_movie"]
        data = {"langs": [lang], "movieName": movie_name}
        response = self.post(url=url, data=data)
        return response

    def get_tvseries(self, lang: str, movie_name: str, season: str, *args, **kwargs) -> requests.Response:
        url = self.method["get_movie"]
        data = {"langs": [lang], "movieName": movie_name, "season": season}
        response = self.post(url=url, data=data)
        return response

    def get_sub(self, movie: str, lang: str, id: str, *args, **kwargs) -> requests.Response:
        url = self.method["get_sub"]
        data = {"movie": movie, "lang": lang, "id": id}
        response = self.post(url=url, data=data)
        return response

    def response_status_ok(self, response: requests.Response) -> bool:
        log.stdout(f"{response.url} status_code: {response.status_code} {response.reason}")
        if response.status_code != 200:
            return False
        return True


class SubsourceParser(common_utils.ProviderHelper):
    def __init__(self, *args, **kwargs) -> None:
        common_utils.ProviderHelper.__init__(self, *args, **kwargs)
        self.api = SubsourceApi()
        self.provider_name = ""
        self.api.provider_name = self.provider_name

    def parse_search_movie_response(self, response: requests.Response) -> dict[str, Any]:
        data = response.json()
        found_items = data["found"]
        parsed_response = {}
        if len(found_items) == 0:
            return {}
        for release in found_items:
            if self.skip_search_movie_item(release):
                continue
            if self.tvseries:
                season = f"season-{self.season_no_padding}"
                parsed_response["season"] = season
            parsed_response["link_name"] = release["linkName"]
        return parsed_response

    def parse_get_movie_response(self, response: requests.Response) -> None:
        data = response.json()
        for sub in data["subs"]:
            if self.skip_get_movie_item(sub):
                continue
            request_data = self.set_post_data(sub)
            subtitle_name = sub["releaseName"]
            if self.skip_tvseries(subtitle_name):
                continue
            self.prepare_subtitle(self.provider_name, subtitle_name, "", request_data=request_data)

    def skip_tvseries(self, subtitle_name: str) -> bool:
        if self.tvseries:
            patterns = self.ok_season_episode_pattern
            return any(item in subtitle_name.lower() for item in patterns)
        return False

    def set_post_data(self, x: dict[str, Any]) -> dict[str, Any]:
        data = {
            "movie": x["linkName"],
            "lang": x["lang"],
            "id": str(x["subId"]),
            "release_name": x["releaseName"],
            "api_method": self.api.method["get_sub"],
        }
        return data

    def release_in_post_data(self, release_name: str, subtitle_data: list[dict[str, Any]]) -> bool:
        for i in subtitle_data:
            if release_name in i.values():
                return True
        return False

    def parse_get_sub_response(self, response: requests.Response) -> str:
        data = response.json()
        sub = data["sub"]

        if self.skip_get_sub_item(sub):
            return ""

        return f"{self.api.method['download_sub']}/{sub['downloadToken']}"

    def skip_search_movie_item(self, items: dict[str, Any]) -> bool:
        keys = ["subCount", "type", "title", "linkName"]
        if not self.keys_exsist(items, keys):
            return True
        if items["subCount"] == 0:
            return True
        elif self.tvseries and items["type"] == "Movie":
            return True
        elif not self.tvseries and items["type"] == "TVSeries":
            return True
        elif items["title"].lower() != self.title:
            return True
        return False

    def skip_get_movie_item(self, items: dict[str, Any]) -> bool:
        keys = ["releaseName", "linkName", "lang", "subId"]
        if not self.keys_exsist(items, keys):
            return True
        elif not self.subtitle_hi_match(items["hi"]):
            return True
        elif not self.subtitle_language_match(items["lang"]):
            return True
        return False

    def skip_get_sub_item(self, items: dict[str, Any]) -> bool:
        keys = ["ri", "downloadToken"]
        if not self.keys_exsist(items, keys):
            return True
        return False


class Subsource(SubsourceParser):
    def __init__(self, *args, **kwargs) -> None:
        SubsourceParser.__init__(self, *args, **kwargs)
        self.provider_name = self.__class__.__name__.lower()

    def start_search(self, *args, **kwargs) -> None:
        self.api.call_limit = 5  # default, requests made in quick succession
        release_results = self.find_release(self.imdb_id)
        self.find_subtitles(release_results)

    def find_release(self, imdb_id: str) -> dict[str, str]:
        response = self.api.search_movie(imdb_id)
        if not self.api.response_status_ok(response):
            return ""
        data = self.parse_search_movie_response(response)
        return data

    def find_subtitles(self, request_data: dict[str, str]) -> None:
        if not request_data:
            return None

        if self.tvseries:
            response = self.api.get_tvseries(self.current_language, request_data["link_name"], request_data["season"])
        else:
            response = self.api.get_movie(self.current_language, request_data["link_name"])

        if not self.api.response_status_ok(response):
            return None

        self.parse_get_movie_response(response)


class GetDownloadUrl:
    def __init__(self) -> None:
        self._api = SubsourceApi()
        self._response = None

    def get_url(self, item: Subtitle) -> str:
        self._response = self._api.get_sub(
            item.request_data["movie"], item.request_data["lang"], item.request_data["id"]
        )
        if self._api.response_status_ok(self._response):
            return self._parse_get_sub_response()
        return ""

    def _parse_get_sub_response(self) -> str:
        data = self._response.json()
        sub = data["sub"]

        if self._skip_get_sub_item(sub):
            return ""

        return f"{self._api.method['download_sub']}/{sub['downloadToken']}"

    def _skip_get_sub_item(self, items: dict[str, Any]) -> bool:
        keys = ["ri", "downloadToken"]
        if not self._keys_exsist(items, keys):
            return True
        return False

    def _keys_exsist(self, dict_: dict[str, Any], keys: list[str]) -> bool:
        for key in keys:
            if key not in dict_:
                return False
        return True



import threading
from typing import Any

import cloudscraper
from cloudscraper import CloudScraper
from requests import Response, exceptions
from selectolax.parser import HTMLParser

from subsearch.globals import log
from subsearch.globals.constants import VIDEO_FILE
from subsearch.globals.dataclasses import (
    AppConfig,
    LanguageData,
    ProviderUrls,
    ReleaseData,
    Subtitle,
)
from subsearch.utils import string_parser

api_calls_made: dict[str, int] = {}
_thread_lock = threading.Lock()


class ProviderDataContainer:
    def __init__(self, *args, **kwargs) -> None:
        release_data: ReleaseData = kwargs.get("release_data")
        app_config: AppConfig = kwargs.get("app_config")
        provider_urls: ProviderUrls = kwargs.get("provider_urls")
        language_data: LanguageData = kwargs.get("language_data")

        self.app_config = app_config
        self.release_data = release_data
        self.provider_urls = provider_urls
        self.language_data = language_data

        # file parameters
        self.title = release_data.title
        self.year = release_data.year
        self.season = release_data.season
        self.season_ordinal = release_data.season_ordinal
        self.episode = release_data.episode
        self.episode_ordinal = release_data.episode_ordinal
        self.tvseries = release_data.tvseries
        self.release = release_data.release
        self.group = release_data.group
        self.imdb_id = release_data.imdb_id

        # user parameters
        self.current_language = app_config.current_language
        self.hearing_impaired = app_config.hearing_impaired
        self.non_hearing_impaired = app_config.non_hearing_impaired
        self.accept_threshold = app_config.accept_threshold
        self.open_on_no_matches = app_config.open_on_no_matches
        self.api_call_limit = app_config.api_call_limit
        self.request_connect_timeout = app_config.request_connect_timeout
        self.request_read_timeout = app_config.request_read_timeout
        self.request_timeout = (self.request_connect_timeout, self.request_read_timeout)
        # provider url data
        self.url_opensubtitles = provider_urls.opensubtitles
        self.url_opensubtitles_hash = provider_urls.opensubtitles_hash
        self.url_yifysubtitles = provider_urls.yifysubtitles
        self.url_subsource = provider_urls.subsource

        self.file_hash = VIDEO_FILE.file_hash
        self.season_no_padding = self._season_no_padding()
        self.episode_no_padding = self._episode_no_padding()
        self.ok_season_episode_pattern = self._ok_season_episode_pattern()

    def _season_no_padding(self) -> str:
        return string_parser.remove_padded_zero(self.release_data.season)

    def _episode_no_padding(self) -> str:
        return string_parser.remove_padded_zero(self.release_data.episode)

    def _ok_season_episode_pattern(self) -> list[str]:
        season_episode_pattern_0 = [
            f"season.{self.season}.",
            f"s{self.season}e{self.episode}.",
            f"s{self.season}.e{self.episode}.",
        ]
        season_episode_pattern_1 = [
            f"season.{self.season_no_padding}.",
            f"s{self.season_no_padding}.e{self.episode_no_padding}.",
        ]
        return season_episode_pattern_0 + season_episode_pattern_1


class _PrepareSubtitleDownload:
    accepted_subtitles: list[Subtitle] = []
    rejected_subtitles: list[Subtitle] = []

    def __init__(self, master: "ProviderHelper", *args, **kwargs) -> None:
        self.provider_name = master.provider_name
        self.subtitle_name = master.subtitle_name
        self.precentage_result = 0
        self.download_url = master.download_url
        self.release_name = VIDEO_FILE.filename
        self.accept_threshold = master.accept_threshold
        self.request_data = master.request_data
        self.master = master

    @property
    def _subtitle(self) -> Subtitle:
        self.subtitle_name = self._sanitize_filename(self.subtitle_name)
        if self.download_url:
            subtitle = self._get_subtitle_no_request_data()
        elif self.request_data:
            subtitle = self._get_subtitle_with_download_url()
        else:
            raise Exception("Subtitle not corectlly populated")
        return subtitle

    def _sanitize_filename(filename: str) -> str:
        old_filename = filename
        invalid_chars = '<>:"/\\|?*'
        for char in invalid_chars:
            filename = filename.replace(char, "_")
        filename = filename.replace("'", "")
        log.stdout(f"Sanitized {old_filename} to {filename}")
        return filename

    def _get_subtitle_no_request_data(self) -> Subtitle:
        subtitle = Subtitle(
            precentage_result=self.precentage_result,
            provider_name=self.provider_name,
            subtitle_name=self.subtitle_name.lower(),
            download_url=self.download_url,
            request_data={},
        )
        return subtitle

    def _get_subtitle_with_download_url(self) -> Subtitle:
        subtitle = Subtitle(
            precentage_result=self.precentage_result,
            provider_name=self.provider_name,
            subtitle_name=self.subtitle_name.lower(),
            download_url="",
            request_data=self.request_data,
        )
        return subtitle

    def prepare_subtitle(self) -> None:
        self.set_percentage_result()
        self.log_subtitle_match()
        self.populate_correct_subtitle_list()

    def set_percentage_result(self) -> None:
        self.precentage_result = string_parser.calculate_match(self.subtitle_name, self.release_name)

    def log_subtitle_match(self) -> None:
        log.subtitle_match(
            self.provider_name,
            self.subtitle_name,
            self.precentage_result,
            self.accept_threshold,
        )

    def populate_correct_subtitle_list(self) -> None:
        if self.master.threshold_met(self.precentage_result):
            return _PrepareSubtitleDownload.accepted_subtitles.append(self._subtitle)
        return _PrepareSubtitleDownload.rejected_subtitles.append(self._subtitle)


class ProviderHelper(ProviderDataContainer):

    def __init__(self, *args, **kwargs) -> None:
        ProviderDataContainer.__init__(self, *args, **kwargs)
        self.provider_name = ""
        self.subtitle_name = ""
        self.download_url = ""

    def _set_subtitle_cls_vars(self, *args, **kwargs) -> None:
        self.provider_name: str = args[0]
        self.subtitle_name: str = args[1]
        self.download_url: str = args[2]
        self.request_data: dict[str, Any] = kwargs.get("request_data", {}) or args[3]

    @property
    def accepted_subtitles(self) -> list[Subtitle]:
        with _thread_lock:
            return _PrepareSubtitleDownload.accepted_subtitles

    @property
    def rejected_subtitles(self) -> list[Subtitle]:
        with _thread_lock:
            return _PrepareSubtitleDownload.rejected_subtitles

    def prepare_subtitle(self, provider_name: str, subtitle_name: str, download_url: str, request_data: dict) -> None:
        with _thread_lock:
            self._set_subtitle_cls_vars(provider_name, subtitle_name, download_url, request_data)
            prepare = _PrepareSubtitleDownload(self)
            prepare.prepare_subtitle()

    def subtitle_hi_match(self, hi: bool) -> bool:
        if self.hearing_impaired and self.non_hearing_impaired:
            return True
        if not self.hearing_impaired and hi:
            return False
        if not self.non_hearing_impaired and hi:
            return False
        return True

    def subtitle_language_match(self, language: str) -> bool:
        if self.current_language.lower() != language.lower():
            return False
        return True

    def keys_exsist(self, dict_: dict[str, Any], keys: list[str]) -> bool:
        for key in keys:
            if key not in dict_:
                return False
        return True

    def threshold_met(self, precentage_result: int) -> bool:
        if precentage_result >= self.accept_threshold:
            return True
        return False


def get_cloudscraper() -> CloudScraper:
    return cloudscraper.create_scraper(browser={"browser": "chrome", "platform": "android", "desktop": False})


def send_request(url: str, scraper: CloudScraper, timeout: tuple[int, int], header=None) -> Response:
    if header is None:
        return scraper.get(url, timeout=timeout)
    return scraper.get(url, timeout=timeout, headers=header)


def parse_scraper_response(response: Response) -> HTMLParser:
    return HTMLParser(response.text)


def request_parsed_response(url: str, timeout: tuple[int, int], header=None) -> HTMLParser:
    scraper = get_cloudscraper()
    try:
        response = send_request(url, scraper, timeout=timeout, header=header)
    except exceptions.Timeout as e:
        log.stdout(e, level="warning", print_allowed=False)
        return ""
    return parse_scraper_response(response)


def sort_list_by_precentage_result(list_: list) -> list:
    with _thread_lock:
        return sorted(list_, key=lambda i: i.precentage_result, reverse=True)



import ctypes
import time
from pathlib import Path
from typing import Callable

from subsearch.globals import decorators, log, thread_handle
from subsearch.globals.constants import APP_PATHS, DEVICE_INFO, FILE_PATHS, VIDEO_FILE
from subsearch.globals.dataclasses import Subtitle
from subsearch.gui import screen_manager, system_tray
from subsearch.gui.screens import download_manager
from subsearch.providers import opensubtitles, subsource, yifysubtitles
from subsearch.utils import imdb_lookup, io_file_system, io_toml, string_parser, io_winreg


class Initializer:
    def __init__(self, pref_counter: float) -> None:
        self.start = pref_counter
        log.brackets("Initializing")

        self.api_calls_made: dict[str, int] = {}
        self.ran_download_tab = False
        self.accepted_subtitles: list[Subtitle] = []
        self.rejected_subtitles: list[Subtitle] = []
        self.manually_accepted_subtitles: list[Subtitle] = []
        self.release_data = string_parser.no_release_data()
        self.provider_urls = string_parser.CreateProviderUrls.no_urls()
        self.file_exist = VIDEO_FILE.file_exist
        self.autoload_src = ""

        self.downloaded_subtitles: int = 0
        self.downloaded_subtitle_archives: int = 0
        self.extracted_subtitle_archives: int = 0
        self.user_downloaded_files = False

        log.stdout("Verifing files and paths", level="debug")
        self.setup_file_system()
        self.language_data = io_toml.load_toml_data(FILE_PATHS.language_data)
        self.app_config = io_toml.get_app_config(FILE_PATHS.config)
        if not io_winreg.check_long_paths_enabled():
            self._notify_user()

        log.dataclass(DEVICE_INFO, level="debug", print_allowed=False)
        log.dataclass(self.app_config, level="debug", print_allowed=False)

        log.stdout("Initializing system tray icon", level="debug")
        decorators.enable_system_tray = self.app_config.system_tray
        self.system_tray = system_tray.SystemTray()
        self.system_tray.start()

        if self.file_exist:
            VIDEO_FILE.file_hash = io_file_system.get_file_hash(VIDEO_FILE.file_path)
            log.dataclass(VIDEO_FILE, level="debug", print_allowed=False)
            io_file_system.create_directory(VIDEO_FILE.file_directory)
            self.release_data = string_parser.get_release_data(VIDEO_FILE.filename)
            self.update_imdb_id()
            log.dataclass(self.release_data, level="debug", print_allowed=False)
            provider_urls = string_parser.CreateProviderUrls(self.app_config, self.release_data, self.language_data)
            self.provider_urls = provider_urls.retrieve_urls()
            log.dataclass(self.provider_urls, level="debug", print_allowed=False)
            self.search_kwargs = dict(
                release_data=self.release_data,
                app_config=self.app_config,
                provider_urls=self.provider_urls,
                language_data=self.language_data,
            )
        self.call_conditions = CallConditions(self)
        log.task_completed()

    def update_imdb_id(self) -> None:
        timeout = self.app_config.request_connect_timeout, self.app_config.request_read_timeout
        find_id = imdb_lookup.FindImdbID(
            self.release_data.title,
            self.release_data.year,
            self.release_data.tvseries,
            request_timeout=timeout,
        )
        self.release_data.imdb_id = find_id.imdb_id

    def setup_file_system(self) -> None:

        io_file_system.create_directory(APP_PATHS.tmp_dir)
        io_file_system.create_directory(APP_PATHS.appdata_subsearch)
        io_toml.resolve_on_integrity_failure()
        io_file_system.del_directory_content(APP_PATHS.tmp_dir)
        if self.file_exist:
            io_file_system.create_directory(VIDEO_FILE.subs_dir)
            io_file_system.create_directory(VIDEO_FILE.tmp_dir)

    def all_providers_disabled(self) -> bool:
        if (
            self.app_config.providers["opensubtitles_site"] is False
            and self.app_config.providers["opensubtitles_hash"] is False
            and self.app_config.providers["yifysubtitles_site"] is False
            and self.app_config.providers["subsource_site"] is False
        ):
            return True
        return False

    def prevent_conflicting_config_settings(self) -> None:
        # TODO
        # make settings exclusive in GUI
        if self.app_config.open_on_no_matches and self.app_config.always_open:
            self.app_config.open_on_no_matches = False
            io_toml.update_toml_key(FILE_PATHS.config, "download_manager.open_on_no_matches", False)
        if (
            self.app_config.subtitle_post_processing["move_best"]
            and self.app_config.subtitle_post_processing["move_all"]
        ):
            self.app_config.subtitle_post_processing["move_best"] = False
            io_toml.update_toml_key(FILE_PATHS.config, "subtitle_post_processing.move_best", False)

    def _notify_user(self) -> None:
        log.stdout("Win32 long paths disabled; paths >260 chars may fail. Set LongPathsEnabled=1 and reboot.")
        log.stdout("https://github.com/vagabondHustler/Win32LongPaths")


class SubsearchCore(Initializer):
    def __init__(self, pref_counter: float) -> None:
        Initializer.__init__(self, pref_counter)
        ctypes.windll.kernel32.SetConsoleTitleW(f"subsearch - {DEVICE_INFO.subsearch}")
        if not self.file_exist:
            log.brackets("GUI")
            screen_manager.open_screen("search_options")
            log.stdout("Exiting GUI", level="debug")
            self.prevent_conflicting_config_settings()
            return None

        if " " in VIDEO_FILE.filename:
            log.stdout(f"{VIDEO_FILE.filename} contains spaces, result may vary", level="warning")

        if not self.all_providers_disabled():
            self.prevent_conflicting_config_settings()
            log.brackets("Search started")

    @decorators.call_func
    def init_search(self, *providers: Callable[..., None]) -> None:
        thread_handle._create_threads(*providers)
        log.task_completed()

    @decorators.call_func
    def opensubtitles(self) -> None:
        thread_handle._start_search(self, provider=opensubtitles.OpenSubtitles, flag="hash")
        thread_handle._start_search(self, provider=opensubtitles.OpenSubtitles, flag="site")

    @decorators.call_func
    def yifysubtitles(self) -> None:
        thread_handle._start_search(self, provider=yifysubtitles.YifiSubtitles, flag="site")

    @decorators.call_func
    def subsource(self) -> None:
        thread_handle._start_search(self, provider=subsource.Subsource, flag="site")

    @decorators.call_func
    def download_files(self) -> None:
        log.brackets(f"Downloading subtitles")
        index_size = len(self.accepted_subtitles)
        self.accepted_subtitles = sorted(self.accepted_subtitles, key=lambda i: i.precentage_result, reverse=True)
        for index_position, subtitle in enumerate(self.accepted_subtitles):
            if subtitle.provider_name not in self.api_calls_made:
                self.api_calls_made[subtitle.provider_name] = 0
            if not self.api_calls_made[subtitle.provider_name] == self.app_config.api_call_limit:
                if subtitle.provider_name == "subsource":
                    self._handle_subsource_subtitle(index_position, subtitle)
                sub_count = sum(self.api_calls_made.values(), 1)
                io_file_system.download_subtitle(subtitle, sub_count, index_size)
                self.downloaded_subtitles += 1
                self.api_calls_made[subtitle.provider_name] += 1
            else:
                s = self.accepted_subtitles.pop(index_position)
                self.rejected_subtitles.append(s)
        log.task_completed()

    def _handle_subsource_subtitle(self, index_position: int, subtitle: Subtitle) -> None:
        if self.app_config.providers["subsource_site"]:
            subsource_api = subsource.GetDownloadUrl()
            download_url = subsource_api.get_url(subtitle)
            if not download_url:
                self.accepted_subtitles.pop(index_position)
                log.stdout(f"{subtitle.provider_name} could not be reached. Removed {subtitle.subtitle_name}")
            else:
                self.accepted_subtitles[index_position].download_url = download_url
                self.accepted_subtitles[index_position].request_data = {}

    @decorators.call_func
    def download_manager(self) -> None:
        log.brackets(f"Download Manager")
        subtitles = self.rejected_subtitles + self.accepted_subtitles
        screen_manager.open_screen("download_manager", subtitles=subtitles)
        self.manually_accepted_subtitles = download_manager.DownloadManager.downloaded_subtitle
        self.downloaded_subtitles += len(self.manually_accepted_subtitles)
        log.task_completed()

    @decorators.call_func
    def subtitle_post_processing(self) -> None:
        target = self.app_config.subtitle_post_processing["target_path"]
        resolution = self.app_config.subtitle_post_processing["path_resolution"]
        target_path = io_file_system.create_path_from_string(target, resolution)
        self.downloaded_subtitle_archives = io_file_system.count_files_in_directory(VIDEO_FILE.tmp_dir)
        self.extract_files()
        self.extracted_subtitle_archives = io_file_system.count_files_in_directory(VIDEO_FILE.subs_dir, [".srt"])
        self.subtitle_rename()
        self.subtitle_move_best(target_path)
        self.subtitle_move_all(target_path)

    @decorators.call_func
    def extract_files(self) -> None:
        log.brackets("Extracting downloads")
        io_file_system.extract_files_in_dir(VIDEO_FILE.tmp_dir, VIDEO_FILE.subs_dir)
        log.task_completed()

    @decorators.call_func
    def subtitle_rename(self) -> None:
        log.brackets("Renaming best match")
        new_name = io_file_system.autoload_rename(VIDEO_FILE.filename, ".srt")
        self.autoload_src = new_name

        log.task_completed()

    @decorators.call_func
    def subtitle_move_best(self, target: Path) -> None:
        log.brackets("Move best match")
        io_file_system.move_and_replace(self.autoload_src, target)
        log.task_completed()

    @decorators.call_func
    def subtitle_move_all(self, target: Path) -> None:
        log.brackets("Move all")
        io_file_system.move_all(VIDEO_FILE.subs_dir, target)
        log.task_completed()

    @decorators.call_func
    def summary_notification(self, elapsed) -> None:
        log.brackets("Summary toast")
        elapsed_summary = f"Finished in {elapsed} seconds"
        number_of_results = len(self.accepted_subtitles) + len(self.rejected_subtitles)
        matches_downloaded = f"Downloaded: {self.downloaded_subtitles}/{number_of_results}"
        if self.downloaded_subtitles >= 1:
            msg = "Search Succeeded", f"{matches_downloaded}\n{elapsed_summary}"
            log.stdout(matches_downloaded, hex_color="#a6e3a1")
            self.system_tray.display_toast(*msg)
        elif self.downloaded_subtitles == 0:
            msg = "Search Failed", f"{matches_downloaded}\n{elapsed_summary}"
            log.stdout(matches_downloaded, hex_color="#f38ba8")
            self.system_tray.display_toast(*msg)

    @decorators.call_func
    def clean_up(self) -> None:
        log.brackets("Cleaning up")
        io_file_system.del_file_type(VIDEO_FILE.subs_dir, ".nfo")
        io_file_system.del_directory_content(APP_PATHS.tmp_dir)
        io_file_system.del_directory(VIDEO_FILE.tmp_dir)
        if io_file_system.directory_is_empty(VIDEO_FILE.subs_dir):
            io_file_system.del_directory(VIDEO_FILE.subs_dir)
        log.task_completed()

    def core_on_exit(self) -> None:
        log.brackets("Exit")
        elapsed = time.perf_counter() - self.start
        self.summary_notification(elapsed)
        self.system_tray.stop()
        log.stdout(f"Finished in {elapsed} seconds", hex_color="#f2cdcd")
        if not self.app_config.show_terminal:
            return None
        if DEVICE_INFO.mode == "executable":
            return None

        try:
            input("Enter to exit")
        except KeyboardInterrupt:
            pass


class CallConditions:

    def __init__(self, cls: SubsearchCore) -> None:
        self.cls = cls
        self.refresh_parent_attrs()

    def refresh_parent_attrs(self) -> None:
        self.app_config = self.cls.app_config
        self.file_exist = self.cls.file_exist
        self.release_data = self.cls.release_data
        self.provider_urls = self.cls.provider_urls
        self.language_data = self.cls.language_data
        self.accepted_subtitles = self.cls.accepted_subtitles
        self.rejected_subtitles = self.cls.rejected_subtitles
        self.downloaded_subtitle_archives = self.cls.downloaded_subtitle_archives
        self.extracted_subtitle_archives = self.cls.extracted_subtitle_archives

    def check_language_compatibility(self, provider: str) -> bool:
        language = self.app_config.current_language
        if provider in self.language_data[language]:
            return False
        return True

    def all_conditions_true(self, conditions: list[bool]) -> bool:
        if False in conditions:
            return False
        return True

    def call_func(self, *args, **kwargs) -> bool:
        func_name = kwargs["func_name"]
        self.refresh_parent_attrs()
        conditions: dict[str, list[bool]] = {
            "init_search": [self.file_exist],
            "opensubtitles": [
                self.file_exist,
                lambda: self.check_language_compatibility("opensubtitles"),
                self.app_config.providers["opensubtitles_hash"] or self.app_config.providers["opensubtitles_site"],
            ],
            "yifysubtitles": [
                self.file_exist,
                not self.app_config.only_foreign_parts,
                lambda: self.check_language_compatibility("yifysubtitles"),
                not self.release_data.tvseries,
                not self.provider_urls.yifysubtitles == "",
                self.app_config.providers["yifysubtitles_site"],
            ],
            "subsource": [
                self.file_exist,
                not self.app_config.only_foreign_parts,
                lambda: self.check_language_compatibility("subsource"),
                self.app_config.providers["subsource_site"],
            ],
            "download_files": [
                (len(self.accepted_subtitles) >= 1 and self.app_config.automatic_downloads)
                or (
                    len(self.accepted_subtitles) >= 1
                    and not self.app_config.automatic_downloads
                    and not self.app_config.always_open
                    and not self.app_config.open_on_no_matches
                ),
            ],
            "download_manager": [
                (
                    (
                        len(self.accepted_subtitles) == 0
                        and len(self.rejected_subtitles) >= 1
                        and self.app_config.open_on_no_matches
                    )
                    or (
                        (len(self.accepted_subtitles) >= 1 or len(self.rejected_subtitles) >= 1)
                        and self.app_config.always_open
                    )
                ),
            ],
            "subtitle_post_processing": [
                self.file_exist,
            ],
            "extract_files": [
                self.downloaded_subtitle_archives >= 1,
            ],
            "subtitle_rename": [
                self.extracted_subtitle_archives >= 1,
                self.app_config.subtitle_post_processing["rename"],
            ],
            "subtitle_move_best": [
                self.extracted_subtitle_archives >= 1,
                self.app_config.subtitle_post_processing["move_best"],
                not self.app_config.subtitle_post_processing["move_all"],
            ],
            "subtitle_move_all": [
                self.extracted_subtitle_archives >= 1,
                self.app_config.subtitle_post_processing["move_all"],
            ],
            "summary_notification": [
                self.file_exist,
                self.app_config.summary_notification,
            ],
            "clean_up": [self.file_exist],
        }

        return self.all_conditions_true(conditions[func_name])
